import h5py, keras, os

import numpy as np
import pandas as pd
import awkward as ak
import tensorflow as tf

from joblib import dump, load

from keras.api import ops

from hepinfo.models.qkerasV3 import quantized_sigmoid

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import roc_curve, auc


def readFromAnomalyh5(inputfile, process, object_ranges='default1', moreInfo=None, verbosity=0):
    """
    Reads data from h5s that either contain multiple signal processes or background that are loaded individually

    inputfile -- h5 file containing either background events or multiple signal processes
    process -- the name of the (signal) process, in case of reading a background file, please set this to 'background'
    object ranges -- dictionary containing the keys "met", "jets", "muons" and "egs" and for each key a tuple containing
                     the range of the corresponding objects in the data array of the h5 file. For "met" it should be given
                     only the index (one number, no tuple), because by definition there can be only one MET object
                     Alternatively it can be set to the string "default1" (4 egs, 4 muons, 10 jets) or
                     "default2" (12 egs, 8 muons, 12 jets)
    moreInfo -- more info about the data that is included in the output
    verbosity -- verbosity of this function
    """

    if verbosity > 0:
        print('Reading anomaly team preprocessed file at ' + inputfile + ' for process ' + process + '.')

    # constructing the information dict
    # some things will be automatically filled here
    # the input "moreInfo" can be used to pass more information
    # this information will have priority over automatically set entries
    infoDict = {}
    infoDict['input'] = inputfile

    # preparing lists to store L1 bit info
    L1bits_labels = []
    L1bits = []

    eventData = {}  # initialize as empty in case file does not contain "event_info" key

    # reading the intput file
    if verbosity > 0:
        print('Starting to read input file...')
    with h5py.File(inputfile, 'r') as h5f2:
        if process == 'background':
            for key in h5f2.keys():
                if key[:3] == 'L1_':
                    L1bits_labels.append(key)
                    L1bits.append(np.array(h5f2[key]))
                elif key == 'L1bit':
                    L1bit = np.array(h5f2[key])
                elif key == 'event_info':
                    # Event info names taken from here:
                    # https://gitlab.cern.ch/cms-l1-ad/l1tntuple-maker/-/blob/master/convert_to_h5.py#L125
                    #                     labels = ["run", "lumi", "event", "bx", "orbit", "time","nPV_True"] #deprecated
                    labels = ['run', 'lumi', 'event', 'bx', 'nPV', 'nPV_Good']
                    index = 0
                    for label in labels:
                        eventData[label] = h5f2[key][:, index]
                        index += 1

                if len(h5f2[key].shape) < 3:
                    continue
                if key == 'full_data_cyl':
                    data = h5f2[key][:, :, :].astype('float')
        else:
            if process not in h5f2.keys():
                raise ValueError(f'The process {process} is not contained in the file {inputfile}.')
            for key in h5f2.keys():
                if key.startswith(process + '_L1_'):
                    L1bits_labels.append(key.replace(process + '_', ''))
                    L1bits.append(np.array(h5f2[key]))
                elif key == process + '_l1bit':
                    L1bit = np.array(h5f2[key])

                # doing this should remove all trigger things, and leave a single entry with the data
                if len(h5f2[key].shape) < 3:
                    continue
                if key == process:
                    data = h5f2[key][:, :, :].astype('float')

    # splitting objects
    if object_ranges == 'default1':
        # we have 57 variables, but they do not have labels yet. Lets assign them based on the info in
        # https://gitlab.cern.ch/cms-l1-ad/l1_anomaly_ae/-/blob/master/in/prep_data.py
        # I assume that we have MET, 4 electrons, 4 muons and 10 jets
        # These are 19 objects, times 3 parameters -> 57 vars
        # From line 27 I think the order is as I listed it: MET, egs, muons, jets
        object_ranges = {'met': 0, 'egs': (1, 5), 'muons': (5, 9), 'jets': (9, 19)}
        if verbosity > 0:
            print(
                'Using the object ranges that are consistent with old h5 files provided by the anomaly detection team containing MET, 4 electrons, 4 muons and 10 jets in this order'
            )
    elif object_ranges == 'default2':
        object_ranges = {'met': 0, 'egs': (1, 13), 'muons': (13, 21), 'jets': (21, 33)}
        if verbosity > 0:
            print(
                'Using the object ranges that are consistent with new h5 files provided by the anomaly detection team containing MET, 12 electrons, 8 muons and 12 jets in this order'
            )

    if verbosity > 0:
        print(f'Object ranges for reading the {process} dataset: {object_ranges}')

    np_sums = data[:, object_ranges['met'], :].reshape((data.shape[0], 1, 3))  # reshape is needed to keep dimensionality
    np_egs = data[:, object_ranges['egs'][0] : object_ranges['egs'][1]]
    np_muons = data[:, object_ranges['muons'][0] : object_ranges['muons'][1]]
    np_jets = data[:, object_ranges['jets'][0] : object_ranges['jets'][1]]

    # converting to awkward
    ak_egs = ak.zip(
        {key: ak.from_regular(np_egs[:, :, i], axis=1) for i, key in enumerate(['pt', 'eta', 'phi'])}, with_name='Momentum4D'
    )
    ak_muons = ak.zip(
        {key: ak.from_regular(np_muons[:, :, i], axis=1) for i, key in enumerate(['pt', 'eta', 'phi'])}, with_name='Momentum4D'
    )
    ak_jets = ak.zip(
        {key: ak.from_regular(np_jets[:, :, i], axis=1) for i, key in enumerate(['pt', 'eta', 'phi'])}, with_name='Momentum4D'
    )

    # energy sums are handled a bit differently
    ak_sums = ak.zip(
        {key: ak.from_regular(np_sums[:, :, 2 * i], axis=1) for i, key in enumerate(['pt', 'phi'])}, with_name='Momentum4D'
    )
    ak_sums['type'] = [2] * len(ak_sums)  # MET should have Type 2

    # removing empty entries (not needed for energy sums)
    ak_egs = ak_egs[ak_egs.pt > 0]
    ak_muons = ak_muons[ak_muons.pt > 0]
    ak_jets = ak_jets[ak_jets.pt > 0]

    infoDict['nEvents'] = len(ak_muons)

    # formating the L1 bits
    d_bits = dict(zip(L1bits_labels, np.asarray(L1bits)))
    d_bits['total L1'] = L1bit
    df_bits = ak.zip(d_bits)

    # after everything else: add moreInfo
    if moreInfo:
        infoDict = {**infoDict, **moreInfo}

    if verbosity > 0:
        print('Done!')

    # we'll output the data as a dict, makes it easier later
    dataDict = {}
    dataDict['muons'] = ak_muons
    dataDict['egs'] = ak_egs
    dataDict['jets'] = ak_jets
    dataDict['sums'] = ak_sums

    return infoDict, eventData, ak.Array(dataDict), df_bits


def readFromAnomalySignalh5(inputfile, process, object_ranges='default1', moreInfo=None, verbosity=0):
    # object ranges -- dictionary containing the keys "met", "jets", "muons" and "egs" and for each key a tuple containing
    #                  the range of the corresponding objects in the data array of the h5 file. For "met" it should be given
    #                  only the index (one number, no tuple), because by definition there can be only one MET object
    #                  Alternatively it can be set to the string "default1" (4 egs, 4 muons, 10 jets) or
    #                  "default2" (12 egs, 8 muons, 12 jets)

    return readFromAnomalyh5(inputfile, process, object_ranges, moreInfo, verbosity)


def readFromAnomalyBackgroundh5(inputfile, object_ranges='default1', moreInfo=None, verbosity=0):
    # object ranges -- dictionary containing the keys "met", "jets", "muons" and "egs" and for each key a tuple containing
    #                  the range of the corresponding objects in the data array of the h5 file. For "met" it should be given
    #                  only the index (one number, no tuple), because by definition there can be only one MET object
    #                  Alternatively it can be set to the string "default1" (4 egs, 4 muons, 10 jets) or
    #                  "default2" (12 egs, 8 muons, 12 jets)

    return readFromAnomalyh5(inputfile, 'background', object_ranges, moreInfo, verbosity)


def awkward_to_numpy(ak_array, maxN, verbosity=0):
    # this is a bit ugly, but it works. Maybe we can improve later
    selected_arr = ak.fill_none(ak.pad_none(ak_array, maxN, clip=True, axis=-1), {'pt': 0, 'eta': 0, 'phi': 0})
    np_arr = np.stack((selected_arr.pt.to_numpy(), selected_arr.eta.to_numpy(), selected_arr.phi.to_numpy()), axis=2)
    return np_arr.reshape(np_arr.shape[0], np_arr.shape[1] * np_arr.shape[2])


def mutual_information_bernoulli_loss(y_true, y_pred, use_quantized_sigmoid=False, bits_bernoulli_sigmoid=8):
    """
    I(x;y)  = H(x)   - H(x|y)
            = H(L_n) - H(L_n|s)
            = H(L_n) - (H(L_n|s=0) + H(L_n|s=1))
    H_bernoulli(x) = -(1-theta) x ln(1-theta) - theta x ln(theta)
    here theta => probability for 1 and 1-theta => probability for 0

    pseudocode:
    def get_h_bernoulli(l):
        theta = np.mean(l, axis=0)
        return -(1-theta) * np.log2(1-theta) - theta * np.log2(theta)

    y_pred = np.random.binomial(n=1, p=0.6, size=[2000, 5])
    y_true = np.random.binomial(n=1, p=0.6, size=[2000])

    y_pred[y_true == 0] = np.random.binomial(n=1, p=0.5, size=[len(y_true[y_true == 0]), 5])
    y_pred[y_true == 1] = np.random.binomial(n=1, p=0.8, size=[len(y_true[y_true == 1]), 5])

    H_L_n = get_h_bernoulli(y_pred)
    H_L_n_s0 = get_h_bernoulli(y_pred[y_true == 0])
    H_L_n_s1 = get_h_bernoulli(y_pred[y_true == 1])

    counts = np.bincount(y_true)

    MI = H_L_n - ((counts[0] / 2000 * H_L_n_s0) + (counts[1] / 2000 * H_L_n_s1))

    return np.sum(MI)

    :param y_pred: output of the layer
    :param y_true: sensitive attribute
    :return: The loss
    """

    def get_h_bernoulli(tensor):
        def get_theta(x):
            alpha = None
            temperature = 6.0
            use_real_sigmoid = use_quantized_sigmoid
            # quantized sigmoid
            _sigmoid = quantized_sigmoid(bits=bits_bernoulli_sigmoid, use_stochastic_rounding=True, symmetric=True)
            if isinstance(alpha, str):
                assert alpha in ['auto', 'auto_po2']

            if isinstance(alpha, str):
                len_axis = len(x.shape)

                if len_axis > 1:
                    if keras.backend.image_data_format() == 'channels_last':
                        axis = list(range(len_axis - 1))
                    else:
                        axis = list(range(1, len_axis))
                else:
                    axis = [0]

                std = ops.std(x, axis=axis, keepdims=True) + keras.backend.epsilon()
            else:
                std = 1.0

            if use_real_sigmoid:
                p = ops.sigmoid(temperature * x / std)
            else:
                p = tf.cast(_sigmoid(temperature * x / std), tf.float64)

            return p

        def log2(x):
            numerator = tf.math.log(x + 1e-20)
            denominator = tf.math.log(tf.constant(2, dtype=numerator.dtype))
            return numerator / denominator

        theta = tf.reduce_mean(get_theta(tensor), axis=0)

        return tf.reduce_sum(-(1 - theta) * log2(1 - theta) - theta * log2(theta))

    def compute_for_value(value, y_true, y_pred, get_h_bernoulli):
        if tf.shape(y_true).shape[0] == 1:
            y_filter = tf.where(y_true == value)
        else:
            y_filter = tf.where(y_true[:, 0] == value)[:, 0]

        y_i = tf.gather(y_pred, indices=y_filter)
        H_L_n_si = get_h_bernoulli(y_i)
        cnt_i = tf.shape(y_i)[0] + tf.cast(1e-16, dtype=tf.int32)  # number of repr with index i
        norm_si = cnt_i / tf.shape(y_pred)[0]
        return H_L_n_si, norm_si

    y_true = tf.cast(y_true, tf.int32)
    y_pred = tf.cast(y_pred, tf.float64)
    H_L_n = get_h_bernoulli(y_pred)

    unique_y_true, _ = tf.unique(y_true)

    H_L_n_s = []
    norm_s = []

    def compute_per_value(v):
        return compute_for_value(v, y_true, y_pred, get_h_bernoulli)

    results = tf.map_fn(compute_per_value, unique_y_true, fn_output_signature=(tf.float64, tf.float64))

    H_L_n_s = tf.convert_to_tensor(results[0])
    norm_s = tf.convert_to_tensor(results[1])

    MI = H_L_n - tf.reduce_sum(tf.math.multiply(norm_s, H_L_n_s))

    # NOTE: this is a hotfix when we dont have all classes
    MI = tf.where(tf.math.is_nan(MI), tf.convert_to_tensor([0.0], dtype=tf.float64), MI)

    return tf.cast(MI, dtype=tf.float32)


class MILoss(keras.losses.Loss):
    def __init__(self, use_quantized_sigmoid=False, bits_bernoulli_sigmoid=8, name="mi_loss", **kwargs):
        super().__init__(name=name, **kwargs)
        self.use_quantized_sigmoid = use_quantized_sigmoid
        self.bits_bernoulli_sigmoid = bits_bernoulli_sigmoid

    def call(self, y_true, y_pred):
        return mutual_information_bernoulli_loss(
            y_true,
            y_pred,
            self.use_quantized_sigmoid,
            self.bits_bernoulli_sigmoid
        )

    def get_config(self):
        return super().get_config()

# following code is based on: https://www.kaggle.com/code/matteomonzali/nn-model

def __rolling_window(data, window_size):
    """
    Rolling window: take window with definite size through the array

    :param data: array-like
    :param window_size: size
    :return: the sequence of windows

    Example: data = array(1, 2, 3, 4, 5, 6), window_size = 4
        Then this function return array(array(1, 2, 3, 4), array(2, 3, 4, 5), array(3, 4, 5, 6))
    """
    shape = data.shape[:-1] + (data.shape[-1] - window_size + 1, window_size)
    strides = data.strides + (data.strides[-1],)
    return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)


def __cvm(subindices, total_events):
    """
    Compute Cramer-von Mises metric.
    Compared two distributions, where first is subset of second one.
    Assuming that second is ordered by ascending

    :param subindices: indices of events which will be associated with the first distribution
    :param total_events: count of events in the second distribution
    :return: cvm metric
    """
    target_distribution = np.arange(1, total_events + 1, dtype='float') / total_events
    subarray_distribution = np.cumsum(np.bincount(subindices, minlength=total_events), dtype='float')
    subarray_distribution /= 1.0 * subarray_distribution[-1]
    return np.mean((target_distribution - subarray_distribution) ** 2)


def compute_cvm(predictions, masses, n_neighbours=200, step=50):
    """
    Computing Cramer-von Mises (cvm) metric on background events: take average of cvms calculated for each mass bin.
    In each mass bin global prediction's cdf is compared to prediction's cdf in mass bin.

    :param predictions: array-like, predictions
    :param masses: array-like, in case of Kaggle tau23mu this is reconstructed mass
    :param n_neighbours: count of neighbours for event to define mass bin
    :param step: step through sorted mass-array to define next center of bin
    :return: average cvm value
    """
    predictions = np.array(predictions)
    masses = np.array(masses)
    assert len(predictions) == len(masses)

    # First, reorder by masses
    predictions = predictions[np.argsort(masses)]

    # Second, replace probabilities with order of probability among other events
    predictions = np.argsort(np.argsort(predictions, kind='mergesort'), kind='mergesort')

    # Now, each window forms a group, and we can compute contribution of each group to CvM
    cvms = []
    for window in __rolling_window(predictions, window_size=n_neighbours)[::step]:
        cvms.append(__cvm(subindices=window, total_events=len(predictions)))
    return np.mean(cvms)


def __roc_curve_splitted(data_zero, data_one, sample_weights_zero, sample_weights_one):
    """
    Compute roc curve

    :param data_zero: 0-labeled data
    :param data_one:  1-labeled data
    :param sample_weights_zero: weights for 0-labeled data
    :param sample_weights_one:  weights for 1-labeled data
    :return: roc curve
    """
    labels = [0] * len(data_zero) + [1] * len(data_one)
    weights = np.concatenate([sample_weights_zero, sample_weights_one])
    data_all = np.concatenate([data_zero, data_one])
    fpr, tpr, _ = roc_curve(labels, data_all, sample_weight=weights)
    return fpr, tpr


def compute_ks(data_prediction, mc_prediction, weights_data, weights_mc):
    """
    Compute Kolmogorov-Smirnov (ks) distance between real data predictions cdf and Monte Carlo one.

    :param data_prediction: array-like, real data predictions
    :param mc_prediction: array-like, Monte Carlo data predictions
    :param weights_data: array-like, real data weights
    :param weights_mc: array-like, Monte Carlo weights
    :return: ks value
    """
    assert len(data_prediction) == len(weights_data), 'Data length and weight one must be the same'
    assert len(mc_prediction) == len(weights_mc), 'Data length and weight one must be the same'

    data_prediction, mc_prediction = np.array(data_prediction), np.array(mc_prediction)
    weights_data, weights_mc = np.array(weights_data), np.array(weights_mc)

    assert np.all(data_prediction >= 0.) and np.all(data_prediction <= 1.), 'Data predictions are out of range [0, 1]'
    assert np.all(mc_prediction >= 0.) and np.all(mc_prediction <= 1.), 'MC predictions are out of range [0, 1]'

    weights_data /= np.sum(weights_data)
    weights_mc /= np.sum(weights_mc)

    fpr, tpr = __roc_curve_splitted(data_prediction, mc_prediction, weights_data, weights_mc)

    Dnm = np.max(np.abs(fpr - tpr))
    return Dnm


def roc_auc_truncated(labels, predictions, tpr_thresholds=(0.2, 0.4, 0.6, 0.8),
                      roc_weights=(4, 3, 2, 1, 0)):
    """
    Compute weighted area under ROC curve.

    :param labels: array-like, true labels
    :param predictions: array-like, predictions
    :param tpr_thresholds: array-like, true positive rate thresholds delimiting the ROC segments
    :param roc_weights: array-like, weights for true positive rate segments
    :return: weighted AUC
    """
    assert np.all(predictions >= 0.) and np.all(predictions <= 1.), 'Data predictions are out of range [0, 1]'
    assert len(tpr_thresholds) + 1 == len(roc_weights), 'Incompatible lengths of thresholds and weights'
    fpr, tpr, _ = roc_curve(labels, predictions)
    area = 0.
    tpr_thresholds = [0.] + list(tpr_thresholds) + [1.]
    for index in range(1, len(tpr_thresholds)):
        tpr_cut = np.minimum(tpr, tpr_thresholds[index])
        tpr_previous = np.minimum(tpr, tpr_thresholds[index - 1])
        area += roc_weights[index - 1] * (auc(fpr, tpr_cut, reorder=True) - auc(fpr, tpr_previous, reorder=True))
    tpr_thresholds = np.array(tpr_thresholds)
    # roc auc normalization to be 1 for an ideal classifier
    area /= np.sum((tpr_thresholds[1:] - tpr_thresholds[:-1]) * np.array(roc_weights))
    return area

def load_tau_data(path):

    if os.path.isfile(f"{path}/x.npy"):
        x = np.load(f"{path}/x.npy")
        y = np.load(f"{path}/y.npy")
        s = np.load(f"{path}/s.npy")
        test = pd.read_pickle(f"{path}/test")
        correlation = pd.read_pickle(f"{path}/correlation")
        agreement_test = pd.read_pickle(f"{path}/agreement_test")

        return x, y, s, test, correlation, agreement_test

    # load data
    # The label ‘signal’ being ‘1’ for signal events, ‘0’ for background events
    # the signal events are simulated, while background events are real data.
    training = pd.read_csv(f"{path}/training.csv")
    # The test dataset has all the columns that training.csv has, except
    # mass, production, min_ANNmuon, and signal.
    test = pd.read_csv(f"{path}/test.csv")
    # Simulated and real events from the channel Ds → φπ to evaluate
    # the performance for simulated-real data
    # It contains the same columns as test.csv and weight column. 
    agreement = pd.read_csv(f"{path}/check_agreement.csv")
    # Only real background events recorded at LHCb to evaluate
    # the mass values locally.
    # It contains the same columns as test.csv and mass column to check correlation
    correlation = pd.read_csv(f"{path}/check_correlation.csv")

    # simulation (source) (signal==1) contains 8205 examples 
    # real data (target) (signal==0) contains 322942 examples with weight values
    counts = np.unique(agreement["signal"], return_counts=True)
    counts_train = np.unique(training["signal"], return_counts=True)

    print(f"Agreement data # source domain: {counts[1][1]}, # target domain: {counts[1][0]}")
    print(f"Training data # source domain: {counts_train[1][1]}, # target domain: {counts_train[1][0]}")

    # prepare the data

    # split the data for agreement test at the end
    agreement_train, agreement_test = \
        train_test_split(agreement, test_size=0.666)

    # we take equal values from both groups
    agreement_train_s0 = agreement_train[agreement_train["signal"]==0].sample(n=int(len(agreement_train[agreement_train["signal"]==1])/2))
    agreement_train_s1 = agreement_train[agreement_train["signal"]==1].sample(n=int(len(agreement_train[agreement_train["signal"]==1])/2))
    agreement = pd.concat([agreement_train_s0, agreement_train_s1])

    # prepare the data for the domain adaptation task
    s = np.concatenate([agreement_train["signal"].to_numpy(), training["signal"].to_numpy()])
    agreement_train = agreement_train.drop(columns=["id", "signal", "SPDhits", "weight"], axis = 1)

    # from the agreement data we only have background so y=0
    y = np.concatenate([np.zeros(len(agreement_train)), training["signal"].to_numpy()])

    # Note: for this example we dont use the mass
    # mass_train = training["mass"]

    # create the features to train on
    x = training.drop(columns=["id", "production", "min_ANNmuon", "signal", "mass", "SPDhits"], axis = 1)
    x = pd.concat([agreement_train, x], ignore_index=True)

    np.save(f"{path}/x", x)
    np.save(f"{path}/y", y)
    np.save(f"{path}/s", s)
    test.to_pickle(f"{path}/test") 
    correlation.to_pickle(f"{path}/correlation") 
    agreement_test.to_pickle(f"{path}/agreement_test")

    return x, y, s, test, correlation, agreement_test


def load_tau_data_split(path):

    if os.path.isfile(f"{path}/X_train.npy"):
        X_train = np.load(f"{path}/X_train.npy")
        X_test = np.load(f"{path}/X_test.npy")
        y_train = np.load(f"{path}/y_train.npy")
        y_test = np.load(f"{path}/y_test.npy")
        S_train = np.load(f"{path}/S_train.npy")
        S_test = np.load(f"{path}/S_test.npy")
        agreement_weight = np.load(f"{path}/agreement_weight.npy")
        agreement_signal = np.load(f"{path}/agreement_signal.npy")
        agreement_test_feature = np.load(f"{path}/agreement_test_feature.npy")
        test = np.load(f"{path}/test.npy")
        correlation = np.load(f"{path}/correlation.npy")
        mass = np.load(f"{path}/mass.npy")
        scaler = load(f"{path}/scaler.joblib")

        train_vali_data = [X_train, X_test, y_train, y_test, S_train, S_test]
        agreement_data = [agreement_weight, agreement_signal, agreement_test_feature]
        meta_data = [test, correlation, scaler, mass]

        return train_vali_data, agreement_data, meta_data

    x, y, s, test, correlation, agreement_test = load_tau_data(path)

    # split the data into train and test set
    X_train, X_test, y_train, y_test, S_train, S_test = \
        train_test_split(x, y, s, test_size=0.333)

    # scale data
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)

    # prepare agreement data
    agreement_weight = agreement_test["weight"]
    agreement_signal = agreement_test["signal"]
    agreement_test_feature = agreement_test.drop(columns=["id", "signal", "SPDhits", "weight"])
    agreement_test_feature = scaler.transform(agreement_test_feature)

    train_vali_data = [X_train, X_test, y_train, y_test, S_train, S_test]
    agreement_data = [agreement_weight, agreement_signal, agreement_test_feature]
    mass = correlation["mass"]
    correlation = correlation.drop(['id', 'mass', 'SPDhits'], axis=1)
    meta_data = [test, correlation, scaler, mass]

    np.save(f"{path}/X_train", X_train)
    np.save(f"{path}/X_test", X_test)
    np.save(f"{path}/y_train", y_train)
    np.save(f"{path}/y_test", y_test)
    np.save(f"{path}/S_train", S_train)
    np.save(f"{path}/S_test", S_test)
    np.save(f"{path}/agreement_weight", agreement_weight)
    np.save(f"{path}/agreement_signal", agreement_signal)
    np.save(f"{path}/agreement_test_feature", agreement_test_feature)
    np.save(f"{path}/test", test)
    np.save(f"{path}/correlation", correlation)
    np.save(f"{path}/mass", mass)
    dump(scaler, f"{path}/scaler.joblib")

    return train_vali_data, agreement_data, meta_data



"""
Class file for the base model
"""

from __future__ import annotations

from typing import Any

import keras
import numpy as np
import pandas as pd

# from nptyping import NDArray
from sklearn.base import BaseEstimator

try:
    from squark.config import QuantizerConfig
    from squark.layers import QDense as SQDense
    from squark.regularizers import MonoL1
except:
    print("WARNING: BaseModel squark not present")

from hepinfo.models.qkerasV3 import QActivation, QDense, BernoulliSampling
from hepinfo.models.QuantFlow import TQActivation, TQDense


class BaseModel(BaseEstimator):
    """
    BaseModel which can hold different wrapper function (_build_model, fit, ...)
    which needs to be implemented in a special model classes
    """

    def __init__(
        self,
        # BaseModel HPs
        hidden_layers: list[int] = [32, 16, 8],
        batch_normalisation_layers: list[int] = [32, 16, 8],  # size and number of layers must match hidden_layers
        quantized_position: list[bool] = [False, True, False],
        batch_normalisation: bool = False,
        activation_binary: str = 'bernoulli',
        activation_nonbinary: str = 'sigmoid',
        acitvation_last_layer: str = 'sigmoid',
        kernel_regularizer: float = 0.0,
        use_s_quark: bool = False,
        use_qkeras: bool = False,
        use_quantflow: bool = False,
        init_quantized_bits=32,
        input_quantized_bits='quantized_bits(16, 6, 0)',
        quantized_bits='quantized_bits(16, 6, 0, use_stochastic_rounding=True)',
        quantized_activation='quantized_relu(10, 6, use_stochastic_rounding=True, negative_slope=0.0)',
        alpha=1,
        beta0=1,
        drop_out: float = 0,
        num_samples: int = 1,
        use_quantized_sigmoid: bool = True,
        bits_bernoulli_sigmoid: int = 8,
        # Common HPs
        batch_size: int = 200,
        learning_rate: float = 0.001,
        learning_rate_decay_rate: float = 1.0,
        learning_rate_decay_steps: int = 1000,
        optimizer: str = 'Adam',
        epoch: int = 10,
        loss: str = 'categorical_crossentropy',
        # other variables
        verbose: int = 0,
        validation_size: float = 0.0,
        input_shape: tuple | int = 0,
        random_seed: int = 42,
        name: str = 'BaseModel',
        dataset_name: str = 'Compas',
        print_summary: bool = False,
        bits: int = 2,
        checkpoint_path: str = '',
        datetime: str = '',
        run_eagerly: bool = False,
        last_layer_size: int = 1,
        conv: bool = False,
    ) -> None:
        # HPs
        self.hidden_layers = hidden_layers
        self.batch_normalisation_layers = batch_normalisation_layers
        self.batch_normalisation = batch_normalisation
        self.activation_binary = activation_binary
        self.activation_nonbinary = activation_nonbinary
        self.acitvation_last_layer = acitvation_last_layer
        self.num_samples = num_samples
        self.use_quantized_sigmoid = use_quantized_sigmoid
        self.bits_bernoulli_sigmoid = bits_bernoulli_sigmoid
        self.kernel_regularizer = kernel_regularizer
        self.init_quantized_bits = init_quantized_bits
        self.input_quantized_bits = input_quantized_bits
        self.quantized_bits = quantized_bits
        self.quantized_activation = quantized_activation
        self.alpha = alpha
        self.beta0 = beta0
        self.use_qkeras = use_qkeras
        self.use_s_quark = use_s_quark
        self.use_quantflow = use_quantflow
        self.drop_out = drop_out
        self.batch_size = batch_size
        if loss == 'binary_crossentropy':
            self.loss = keras.losses.BinaryCrossentropy(from_logits=False)
        elif loss == 'sparse_categorical_crossentropy':
            self.loss = keras.losses.SparseCategoricalCrossentropy(from_logits=False)
        else:
            self.loss = loss
        self.learning_rate = learning_rate
        self.learning_rate_decay_rate = learning_rate_decay_rate
        self.learning_rate_decay_steps = learning_rate_decay_steps
        if optimizer == 'Adam':
            self.optimizer = keras.optimizers.Adam
        if optimizer == 'SGD':
            self.optimizer = keras.optimizers.SGD
        if optimizer == 'Nadam':
            raise NotImplementedError
        self.quantized_position = quantized_position
        self.optimizer_name = optimizer
        self.epoch = epoch
        self.bits = bits
        self.checkpoint_path = checkpoint_path + '/' + datetime + '/'
        self.datetime = datetime
        self.run_eagerly = run_eagerly

        # other variables
        self.model = Any
        self.verbose = verbose
        self.validation_size = validation_size
        self.input_shape = input_shape
        self.random_seed = random_seed
        self.name = name
        self.dataset_name = dataset_name
        self.print_summary = print_summary
        self.last_layer_size = last_layer_size
        self.conv = conv

    def _plot_model(self, model: Any, output_name: str) -> None:
        """
        Plot the model and save it to png file using
        keras.utils.plot_model

        Args:
            Any: model
            str: output_name

        Returns:
            None
        """

        keras.utils.plot_model(
            model,
            to_file=output_name,
            show_shapes=False,
            show_dtype=False,
            show_layer_names=True,
            rankdir='TB',
            expand_nested=False,
            dpi=96,
            layer_range=None,
            show_layer_activations=False,
        )

    def _get_quantized_activation(self, activation_str: str) -> str:
        """
        Generates the string for the binary activation using
        self.bits for the number of bits.

        Args:
            str: activation_str

        Returns:
            str: quantized_activation
        """

        if activation_str == 'tanh':
            return f'quantized_tanh({self.bits}, symmetric=1)'
        if activation_str == 'sigmoid':
            # TODO: check for symmetric=1
            return f'quantized_sigmoid({self.bits})'
        if activation_str == 'bernoulli':
            return 'bernoulli'
        raise ValueError(f'Activation string {activation_str} not recognized')

    def _get_cls_part(
        self,
        input_layer: Any,
        num_relevance_classes: int = 2,
        feature_activation: str = 'softmax',
        kernel_regularizer: Any = None,
        name: str = 'cls_part',
        index: int = 0,
    ) -> Any:
        """
        Build all the output layer of the model.

        Args:
            Any: input_layer
            int: num_relevance_classes
            str: feature_activation
            Any: kernel_regularizer
            str: name
            int: index

        Returns:
            Any: output_layer
        """

        if self.use_qkeras:
            out = QDense(
                num_relevance_classes,
                kernel_initializer='glorot_uniform',
                kernel_quantizer=self.quantized_bits,
                bias_quantizer=self.quantized_bits,
                activation=feature_activation,
                name=f'{name}_{index}',
            )(input_layer)
        elif self.use_quantflow:
            out = TQDense(
                num_relevance_classes,
                init_bits=self.init_quantized_bits,
                activation=feature_activation,
                alpha=self.alpha,
                name=f'{name}_{index}',
            )(input_layer)
        elif self.use_s_quark:
            out = SQDense(num_relevance_classes, activation=feature_activation, beta0=self.beta0, name=f'{name}_{index}')(
                input_layer
            )
        else:
            out = keras.layers.Dense(
                units=num_relevance_classes,
                activation=feature_activation,
                kernel_regularizer=kernel_regularizer,
                activity_regularizer=kernel_regularizer,
                name=f'{name}_{index}',
            )(input_layer)

        return out

    def _get_hidden_qlayer(
        self,
        input_layer: Any,
        hidden_layer: list[int] = [10, 5],
        drop_out: float = 0.0,
        kernel_regularizer: Any = None,
        name: str = '',
        conv: bool = False,
        kernel_size: tuple = (3, 3),
    ) -> Any:
        """
        Build all the hidden layers of the model.

        Args:
            Any: input_layer
            list[int]: hidden_layer
            float: drop_out
            Any: kernel_regularizer
            str: name
            conv: True if you want a convolutional net. The filter numbers will be taken from
            hidden_layer.
            kernel_size: the filter size for convnets. Ignored if conv = False.

        Returns:
            Any: output_layer
        """

        if self.use_qkeras:
            hidden_layers = QActivation(self.input_quantized_bits)(input_layer)
        if self.use_quantflow:
            hidden_layers = TQActivation(self.input_shape, bits=self.init_quantized_bits, alpha=self.alpha)(input_layer)
        if self.use_s_quark:
            hidden_layers = input_layer
            iq_conf = QuantizerConfig(place='datalane', k0=1)
            oq_conf = QuantizerConfig(place='datalane', k0=1, fr=MonoL1(1e-3))

        if not self.use_qkeras and not self.use_quantflow and not self.use_s_quark:
            hidden_layers = input_layer

        last_quantized = None
        # loop over the number of hidden layers for the whole network
        for i, layer in enumerate(self.hidden_layers):
            if self.batch_normalisation and i != 0:
                hidden_layers = keras.layers.BatchNormalization()(hidden_layers)
            if conv:
                hidden_layers = keras.layers.Conv2D(
                    filters=layer,
                    kernel_size=kernel_size,
                    activation=self.activation_nonbinary,
                    kernel_regularizer=kernel_regularizer,
                    activity_regularizer=kernel_regularizer,
                    name=f'{name}_{i}',
                )(hidden_layers)
            else:
                if self.use_qkeras:
                    hidden_layers = QDense(
                        layer,
                        kernel_initializer='glorot_uniform',
                        kernel_quantizer=self.quantized_bits,
                        bias_quantizer=self.quantized_bits,
                        activation=self.quantized_activation,
                        name=f'{name}_{i}',
                    )(hidden_layers)
                elif self.use_quantflow:
                    hidden_layers = TQDense(
                        layer,
                        init_bits=self.init_quantized_bits,
                        activation=self.activation_nonbinary,
                        alpha=self.alpha,
                        name=f'{name}_{i}',
                    )(hidden_layers)
                elif self.use_s_quark:
                    if i == 0:
                        hidden_layers = SQDense(
                            layer, activation=self.activation_nonbinary, iq_conf=iq_conf, beta0=self.beta0, name=f'{name}_{i}'
                        )(hidden_layers)
                    elif i == len(self.hidden_layers) - 1:
                        hidden_layers = SQDense(
                            layer,
                            activation=self.activation_nonbinary,
                            oq_conf=oq_conf,
                            enable_oq=False,
                            beta0=self.beta0,
                            name=f'{name}_{i}',
                        )(hidden_layers)
                    else:
                        hidden_layers = SQDense(
                            layer, activation=self.activation_nonbinary, beta0=self.beta0, name=f'{name}_{i}'
                        )(hidden_layers)
                else:
                    hidden_layers = keras.layers.Dense(
                        units=layer,
                        activation=self.activation_nonbinary,
                        kernel_regularizer=kernel_regularizer,
                        activity_regularizer=kernel_regularizer,
                        name=f'{name}_{i}',
                    )(hidden_layers)
            if self.quantized_position[i]:
                last_quantized = hidden_layers
                hidden_layers = BernoulliSampling(
                    self.num_samples,
                    use_quantized=self.use_quantized_sigmoid,
                    bits_bernoulli_sigmoid=self.bits_bernoulli_sigmoid,
                    name=f'qact_{name}_{i}',
                )(hidden_layers)

            if drop_out > 0:
                hidden_layers = keras.layers.Dropout(drop_out)(hidden_layers)

        # flatten if net is convolutional
        if conv:
            hidden_layers = keras.layers.Flatten()(hidden_layers)

        if self.batch_normalisation:
            hidden_layers = keras.layers.BatchNormalization()(hidden_layers)

        # create classification part
        output_layer = self._get_cls_part(
            input_layer=hidden_layers,
            num_relevance_classes=self.last_layer_size,
            feature_activation=self.acitvation_last_layer,
            kernel_regularizer=keras.regularizers.l2(self.kernel_regularizer),
            name=name,
            index=len(hidden_layer),
        )

        return output_layer, last_quantized

    def _get_hidden_layer(
        self,
        input_layer,
        hidden_layer=[10, 5],
        drop_out=0,
        feature_activation='tanh',
        last_activation='',
        reg=0,
        name='',
        build_bias=False,
        hybrid=False,
        bits=2,
        qactivation='bernoulli',
        qLayer=10,
        conv=False,
    ):
        nn = input_layer
        for i in range(len(hidden_layer)):
            if not conv:
                nn = keras.layers.Dense(
                    units=hidden_layer[i],
                    activation=feature_activation,
                    kernel_regularizer=keras.regularizers.l2(reg),
                    bias_regularizer=keras.regularizers.l2(reg),
                    activity_regularizer=keras.regularizers.l2(reg),
                    name=f'nn_{name}_{i}',
                )(nn)
            else:
                nn = keras.layers.Conv2D(
                    hidden_layer[i],
                    kernel_size=(3, 3),
                    activation='relu',
                    kernel_regularizer=keras.regularizers.l2(reg),
                    bias_regularizer=keras.regularizers.l2(reg),
                    activity_regularizer=keras.regularizers.l2(reg),
                    name=f'conv_{name}_{i}',
                )(nn)
                nn = keras.layers.MaxPooling2D()(nn)

            if drop_out > 0 and (i < (len(hidden_layer) - 1) or last_activation != ''):
                nn = keras.layers.Dropout(drop_out)(nn)

        if last_activation != '':
            nn = keras.layers.Dense(
                units=1,
                activation=last_activation,
                kernel_regularizer=keras.regularizers.l2(reg),
                bias_regularizer=keras.regularizers.l2(reg),
                activity_regularizer=keras.regularizers.l2(reg),
                name=f'nn_out_{name}',
            )(nn)

        if conv:
            nn = keras.layers.Flatten(name='flatten')(nn)

        if build_bias:
            hidden_part = nn
        else:
            hidden_part = keras.models.Model(inputs=input_layer, outputs=nn, name=name)

        if self.verbose == 2 and not build_bias:
            hidden_part.summary()

        return hidden_part

    def _get_ranking_part(self, input_layer, units=1, feature_activation='tanh', reg=0, use_bias=False, name='ranking_part'):
        out = keras.layers.Dense(
            units=units,
            activation=feature_activation,
            use_bias=use_bias,
            kernel_regularizer=keras.regularizers.l2(reg),
            activity_regularizer=keras.regularizers.l2(reg),
            name=name,
        )(input_layer)

        return out

    def _build_model(self) -> None:
        """
        Wrapper function which needs to be implemented in a special
        model class.

        Args:
            NoInput
        Returns:
            None
        """

        raise NotImplementedError

    def fit(self, x_train: np.ndarray, y_train: np.ndarray, **fit_params: dict[Any, Any]) -> None:
        """
        Wrapper function which needs to be implemented in a special
        model class.

        Args:
            NDArray: x_train
            NDArray: y_train
            dict[Any]: fit_params

        Returns:
            None
        """

        raise NotImplementedError

    def predict_proba(self, features: np.ndarray) -> np.ndarray:
        """
        Get the class probablities.

        Args:
            NDArray: features
        Returns:
            NDArray: class probablities
        """

        if len(features.shape) == 1:
            features = [features]  # type: ignore

        res = self.model.predict([features], batch_size=self.batch_size, verbose=self.verbose)[0]  # type: ignore

        return res

    def predict(self, features: np.ndarray, threshold: np.ndarray) -> list[int]:
        """
        Predict a binary class value with a threshold.

        Args:
            NDArray: features
            int: threshold
        Returns:
            list[int]: predictions
        """

        if len(features.shape) == 1:
            features = [features]  # type: ignore

        res = self.model.predict([features], batch_size=self.batch_size, verbose=self.verbose)  # type: ignore

        return [1 if r > threshold else 0 for r in res[0]]

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of the object while dropping the tensorflow stuff.
        Useful to keep track of hyperparameters at the experiment level.

        Args:
            None
        Returns:
            dict[str]: dict of class attributes
        """

        attr_dict = dict(vars(self))

        for key in ['optimizer']:
            try:
                attr_dict.pop(key)
            except KeyError:
                pass

        return attr_dict

    def get_complexity(self) -> int:
        """
        Returns the number of trainable weights of the model.

        Args:
            None
        Returns:
            int: number of trainable weights
        """

        return int(np.sum([keras.backend.count_params(w) for w in self.model.trainable_weights]))  # type: ignore

    def convert_array_to_float(self, array):
        if isinstance(array, np.ndarray):
            array = array.astype('float32')
        if isinstance(array, pd.DataFrame):
            array = array.values.astype('float32')
        return array



import gc

import keras
import numpy as np
import tensorflow as tf

from hepinfo.models.BaseModel import *
from hepinfo.models.GradReverse import GradReverse


class DebiasClassifier(BaseModel):
    """
    TODO
    """

    def __init__(
        self,
        # DebiasClassifier HPs
        hidden_layers=[1024, 512, 128],
        bias_layers=[1024, 512, 128],
        feature_activation='tanh',
        kernel_regularizer=0.0,
        drop_out=0,
        drop_out_bias=0,
        gamma=1.0,
        # Common HPs
        batch_size=200,
        learning_rate=0.001,
        learning_rate_decay_rate=1,
        learning_rate_decay_steps=1000,
        optimizer='Adam',  # 'Nadam' 'SGD'
        epoch=10,
        loss='binary_crossentropy',
        fair_loss='binary_crossentropy',
        # other variables
        verbose=0,
        print_summary=False,
        validation_size=0.1,
        random_seed=42,
        name='DebiasClassifier',
        dataset_name='Compas',
        datetime='',
        run_eagerly=False,
        conv=False,
    ):
        super().__init__(
            # DebiasClassifier HPs
            hidden_layers=hidden_layers,
            kernel_regularizer=kernel_regularizer,
            drop_out=drop_out,
            # Common HPs
            batch_size=batch_size,
            learning_rate=learning_rate,
            learning_rate_decay_rate=learning_rate_decay_rate,
            learning_rate_decay_steps=learning_rate_decay_steps,
            optimizer=optimizer,
            epoch=epoch,
            loss=loss,
            # other variables
            verbose=verbose,
            print_summary=print_summary,
            validation_size=validation_size,
            random_seed=random_seed,
            name=name,
            dataset_name=dataset_name,
            datetime=datetime,
            run_eagerly=run_eagerly,
            conv=conv,
        )

        # DebiasClassifier HPs
        self.bias_layers = bias_layers
        self.drop_out_bias = drop_out_bias
        self.last_layer_size = 1
        self.feature_activation = feature_activation
        self.gamma = gamma
        self.fair_loss = fair_loss

    def _build_model(self):
        """
        TODO
        """

        # Placeholders for the inputs
        self.x0 = keras.layers.Input(shape=self.num_features, name='x0')

        input_layer = keras.layers.Input(shape=self.num_features, name='input')

        self.feature_part = self._get_hidden_layer(
            input_layer=input_layer,
            hidden_layer=self.hidden_layers,
            drop_out=self.drop_out,
            feature_activation=self.feature_activation,
            last_activation='',
            reg=self.kernel_regularizer,
            name='nn',
        )

        if self.print_summary:
            self.feature_part.summary()
            self._plot_model(self.feature_part, 'feature_part.png')

        # Cls Part
        nn0 = self.feature_part(self.x0)

        # Debias Layers
        nn_bias_0 = GradReverse()(nn0)

        nn_bias_0 = self._get_hidden_layer(
            input_layer=nn_bias_0,
            hidden_layer=self.bias_layers,
            drop_out=self.drop_out_bias,
            feature_activation=self.feature_activation,
            last_activation='sigmoid',
            reg=self.kernel_regularizer,
            name='bias',
            build_bias=True,
        )

        out = self._get_ranking_part(
            input_layer=nn0,
            units=self.last_layer_size,
            feature_activation='sigmoid',
            reg=self.kernel_regularizer,
            use_bias=True,
            name='cls_part',
        )

        self.model = keras.models.Model(inputs=self.x0, outputs=[out, nn_bias_0], name=self.name)

        if self.print_summary:
            self.model.summary()
            self._plot_model(self.model, 'model.png')

        lr_schedule = keras.optimizers.schedules.InverseTimeDecay(
            self.learning_rate,
            decay_steps=self.learning_rate_decay_steps,
            decay_rate=self.learning_rate_decay_rate,
            staircase=False,
        )

        losses = {
            'cls_part': self.loss,
            'nn_out_bias': self.fair_loss,
        }

        lossWeights = {'cls_part': 1, 'nn_out_bias': float(self.gamma)}
        metrics = {'cls_part': 'AUC' if self.last_layer_size == 1 else 'acc', 'nn_out_bias': 'acc'}

        self.model.compile(
            optimizer=self.optimizer(lr_schedule),
            loss=losses,
            loss_weights=lossWeights,
            metrics=metrics,
            run_eagerly=self.run_eagerly,
        )

    def fit(self, x, y, s, **fit_params):
        """
        TODO
        x: numpy array of shape [num_instances, num_features]
        y: numpy array of shape [num_instances, last_layer_size]
        s: numpy array of shape [num_instances, 1]
        """
        # set seed
        tf.random.set_seed(self.random_seed)

        # get the correct numpy type
        x = self.convert_array_to_float(x)
        y = self.convert_array_to_float(y)
        s = self.convert_array_to_float(s)

        self.num_features = (x.shape[1],)

        # convert for cls but not for external rankers
        if len(np.unique(y)) != 2:
            self.last_layer_size = len(np.unique(y))
            self.loss = 'categorical_crossentropy'
            y = self.one_hot_convert(y, self.last_layer_size)

        self._build_model()

        ydict = {'cls_part': y, 'nn_out_bias': s}

        self.history = self.model.fit(
            x=x,
            y=ydict,
            batch_size=self.batch_size,
            epochs=self.epoch,
            verbose=self.verbose,
            shuffle=True,
            validation_split=self.validation_size,
        )
        # https://github.com/tensorflow/tensorflow/issues/14181
        # https://github.com/tensorflow/tensorflow/issues/30324
        gc.collect()

        return self.history

    def one_hot_convert(self, y, num_classes):
        arr = np.zeros((len(y), num_classes))
        for i, yi in enumerate(y):
            arr[i, int(yi) - 1] = 1
        return arr



from __future__ import annotations

import gc
from typing import Any
from functools import partial

import keras
from keras.api.saving import register_keras_serializable

import numpy as np

try:
    from squark.utils.sugar import FreeEBOPs
except:
    print("WARNING: BinaryMI squark not present")

from hepinfo.models.BaseModel import BaseModel
from hepinfo.models.QuantFlow import FreeBOPs
from hepinfo.util import MILoss


class BinaryMI(BaseModel):
    """
    Stochastically Quantized Neural Network which has a bernoulli
    activation after each layer to exactly compute the mutual information.
    """

    def __init__(
        self,
        # BinaryMI HPs
        hidden_layers: list[int] = [1024, 512, 128],
        batch_normalisation_layers: list[int] = [1024, 512, 128],
        quantized_position: list[bool] = [False, True, False],
        batch_normalisation: bool = False,
        activation_binary: str = 'bernoulli',
        activation_nonbinary: str = 'tanh',
        acitvation_last_layer: str = 'sigmoid',
        kernel_regularizer: float = 0.01,
        drop_out: float = 0.0,
        gamma: float = 0.0,
        use_s_quark: bool = False,
        use_qkeras: bool = False,
        use_quantflow: bool = False,
        init_quantized_bits=32,
        input_quantized_bits='quantized_bits(16, 6, 0)',
        quantized_bits='quantized_bits(16, 6, 0, use_stochastic_rounding=True)',
        quantized_activation='quantized_relu(10, 6, use_stochastic_rounding=True, negative_slope=0.0)',
        alpha=1,
        beta0=1,
        num_samples=1,
        use_quantized_sigmoid=True,
        bits_bernoulli_sigmoid=8,
        # Common HPs
        batch_size: int = 200,
        learning_rate: float = 0.001,
        learning_rate_decay_rate: float = 1,
        learning_rate_decay_steps: int = 1000,
        optimizer: str = 'Adam',
        epoch: int = 10,
        loss: str = 'binary_crossentropy',
        run_eagerly: bool = False,
        # other variables
        verbose: int = 0,
        validation_size: float = 0.1,
        input_shape: tuple | int = 0,
        last_layer_size: int = 1,
        random_seed: int = 42,
        name: str = 'BinaryMI',
        dataset_name: str = 'Mnist',
        print_summary: bool = False,
        bits: int = 2,
        checkpoint_path: str = '',
        datetime: str = '',
        conv: bool = False,
    ) -> None:
        super().__init__(
            # DirectRankerAdv HPs
            hidden_layers=hidden_layers,
            batch_normalisation_layers=batch_normalisation_layers,
            quantized_position=quantized_position,
            batch_normalisation=batch_normalisation,
            activation_binary=activation_binary,
            activation_nonbinary=activation_nonbinary,
            acitvation_last_layer=acitvation_last_layer,
            kernel_regularizer=kernel_regularizer,
            drop_out=drop_out,
            use_s_quark=use_s_quark,
            use_qkeras=use_qkeras,
            use_quantflow=use_quantflow,
            init_quantized_bits=init_quantized_bits,
            input_quantized_bits=input_quantized_bits,
            quantized_bits=quantized_bits,
            quantized_activation=quantized_activation,
            alpha=alpha,
            beta0=beta0,
            num_samples=num_samples,
            use_quantized_sigmoid=use_quantized_sigmoid,
            bits_bernoulli_sigmoid=bits_bernoulli_sigmoid,
            # Common HPs
            batch_size=batch_size,
            learning_rate=learning_rate,
            learning_rate_decay_rate=learning_rate_decay_rate,
            learning_rate_decay_steps=learning_rate_decay_steps,
            optimizer=optimizer,
            epoch=epoch,
            loss=loss,
            # other variables
            verbose=verbose,
            validation_size=validation_size,
            input_shape=input_shape,
            random_seed=random_seed,
            name=name,
            dataset_name=dataset_name,
            print_summary=print_summary,
            bits=bits,
            checkpoint_path=checkpoint_path,
            datetime=datetime,
            run_eagerly=run_eagerly,
            last_layer_size=last_layer_size,
            conv=conv,
        )

        self.x_input = Any
        self.out = Any
        self.gamma = gamma

    def _build_model(self) -> None:
        r"""
        Is building the model by sequential adding layers.

        1. The data (x) gets into an input layer which is activated
        with a bernoulli layer via a sigmoid:

        .. math::
            \hat{x} = \mathrm{Bern} (\sigma(6.0 \cdot x / 1.0))

        2. The size of the next layers are defined with self.hidden_layers.
        The basic architecture is build via:

        .. code-block:: python

            for i in self.hidden_layers:
                layer = Dense()(layer)
                layer = QActivation('bernoulli')(layer)

        3. To avoid overfitting it is often useful to lower the learning rate
        during the training. Therefore, the model is build using a schedule
        which applies the inverse decay function to an initial learning rate.

        4. Different optimizers can be used such as Adam, Nadam or SGD.

        Args:
            NoInput
        Returns:
            None
        """

        # placeholders for the inputs: shape depends on whether we have a convnet or not
        self.x_input = keras.layers.Input(shape=self.input_shape, name='x')

        # build layers layers
        self.out, self.last_quantized = self._get_hidden_qlayer(
            self.x_input,
            hidden_layer=self.hidden_layers,
            drop_out=self.drop_out,
            name='t',
            kernel_regularizer=keras.regularizers.l2(self.kernel_regularizer),
            conv=self.conv,
        )

        if sum(self.quantized_position) > 0:
            outputs = [self.out, self.last_quantized]
            self.index_qact = [i for i, x in enumerate(self.quantized_position) if x][-1]
            loss = {f't_{len(self.hidden_layers)}': self.loss, f't_{self.index_qact}': MILoss(self.use_quantized_sigmoid, self.bits_bernoulli_sigmoid)}
            lossWeights = {f't_{len(self.hidden_layers)}': 1, f't_{self.index_qact}': float(self.gamma)}
            metrics = {
                f't_{len(self.hidden_layers)}': 'AUC' if self.last_layer_size == 1 else 'acc',
                f't_{self.index_qact}': 'acc',
            }
        else:
            outputs = self.out
            loss = [self.loss]
            lossWeights = float(1)
            metrics = ['AUC'] if self.last_layer_size == 1 else ['acc']

        # create the model
        self.model = keras.models.Model(inputs=self.x_input, outputs=outputs, name=self.name)

        if self.print_summary:
            self.model.summary()  # type: ignore

        # setup learning rate schedule
        # TODO: maybe we have to go with a simple learning rate here for the experiments
        lr_schedule = keras.optimizers.schedules.InverseTimeDecay(
            self.learning_rate,
            decay_steps=self.learning_rate_decay_steps,
            decay_rate=self.learning_rate_decay_rate,
            staircase=False,
        )

        if sum(self.quantized_position) > 0:
            self.model.compile(  # type: ignore
                optimizer=self.optimizer(lr_schedule),
                loss=loss,
                loss_weights=lossWeights,
                metrics=metrics,
                run_eagerly=self.run_eagerly,
            )
        else:
            self.model.compile(  # type: ignore
                optimizer=self.optimizer(lr_schedule), loss=loss, metrics=metrics, run_eagerly=self.run_eagerly
            )

    def fit(self, x_train: np.ndarray, y_train: np.ndarray, s_train: np.ndarray, **fit_params: dict[Any, Any]) -> None:
        """
        Fit function which first build the model and than
        fits it using x_train and y_train. A special callback
        class is used to compute the mutual information at the
        beginning of each epoch and at the end of the training.

        Args:
            NDArray: x_train
            NDArray: y_train
            NDArray: s_train
            dict[Any]: fit_params

        Returns:
            None
        """

        # get the correct numpy type
        x_train = self.convert_array_to_float(x_train)
        y_train = self.convert_array_to_float(y_train)
        s_train = self.convert_array_to_float(s_train)

        # input size
        self.input_shape = (x_train.shape[1],)

        # convert for classifier output
        if self.last_layer_size > 1:
            y_train = keras.utils.to_categorical(y_train, self.last_layer_size)

        self._build_model()

        callback = []
        if self.use_s_quark:
            nan_terminate = keras.callbacks.TerminateOnNaN()
            ebops = FreeEBOPs()
            callback.append(nan_terminate)
            callback.append(ebops)

        if self.use_quantflow:
            nan_terminate = keras.callbacks.TerminateOnNaN()
            bops = FreeBOPs()
            callback.append(nan_terminate)
            callback.append(bops)

        if sum(self.quantized_position) > 0:
            y_train = {f't_{len(self.hidden_layers)}': y_train, f't_{self.index_qact}': s_train}

        history = self.model.fit(  # type: ignore
            x=x_train,
            y=y_train,
            batch_size=self.batch_size,
            epochs=self.epoch,
            verbose=self.verbose,
            shuffle=True,
            validation_split=self.validation_size,
            callbacks=callback,
        )
        # https://github.com/tensorflow/tensorflow/issues/14181
        # https://github.com/tensorflow/tensorflow/issues/30324
        gc.collect()

        return history

    def predict_proba(self, features: np.ndarray) -> np.ndarray:
        """
        Get the class probablities.

        Args:
            NDArray: features
        Returns:
            NDArray: class probablities
        """

        if len(features.shape) == 1:
            features = [features]  # type: ignore

        res = self.model.predict(  # type: ignore
            features, batch_size=self.batch_size, verbose=str(self.verbose)
        )[0]

        return res



### qkeras layers which use kerasV3 ###

import logging
import re
import warnings

import keras
import numpy as np
import tensorflow as tf
from keras.api import activations, constraints, initializers, ops, regularizers
from pyparsing import Group, Optional, Regex, Suppress, delimitedList


def _create_variable_name(attr_name, var_name=None):
    """Creates variable name.
    Arguments:
      attr_name: string. attribute name
      var_name: string. variable name

    Returns:
      string. variable name
    """

    if var_name:
        return var_name + '/' + attr_name

    # This naming scheme is to solve a problem of a layer having more than
    # one quantizer can have multiple qnoise_factor variables with the same
    # name of "qnoise_factor".
    return attr_name + '_' + str(ops.get_uid(attr_name))


class BaseQuantizer(tf.Module):
    """Base quantizer

    Defines behavior all quantizers should follow.
    """

    def __init__(self):
        self.built = False

    def build(self, var_name=None, use_variables=False):
        if use_variables:
            if hasattr(self, 'qnoise_factor'):
                self.qnoise_factor = tf.Variable(
                    lambda: tf.constant(self.qnoise_factor, dtype=tf.float32),
                    name=_create_variable_name('qnoise_factor', var_name=var_name),
                    dtype=tf.float32,
                    trainable=False,
                )
            if hasattr(self, 'integer'):
                self.integer = tf.Variable(
                    lambda: tf.constant(self.integer, dtype=tf.int32),
                    name=_create_variable_name('integer', var_name=var_name),
                    dtype=tf.int32,
                    trainable=False,
                )
        self.built = True

    def _set_trainable_parameter(self):
        pass

    def update_qnoise_factor(self, qnoise_factor):
        """Update qnoise_factor."""
        if isinstance(self.qnoise_factor, tf.Variable):
            # self.qnoise_factor is a tf.Variable.
            # This is to update self.qnoise_factor during training.
            self.qnoise_factor.assign(qnoise_factor)
        else:
            if isinstance(qnoise_factor, tf.Variable):
                # self.qnoise_factor is a numpy variable, and qnoise_factor is a
                # tf.Variable.
                self.qnoise_factor = qnoise_factor.eval()
            else:
                # self.qnoise_factor and qnoise_factor are numpy variables.
                # This is to set self.qnoise_factor before building
                # (creating tf.Variable) it.
                self.qnoise_factor = qnoise_factor

    # Override not to expose the quantizer variables.
    @property
    def variables(self):
        return ()

    # Override not to expose the quantizer variables.
    @property
    def trainable_variables(self):
        return ()

    # Override not to expose the quantizer variables.
    @property
    def non_trainable_variables(self):
        return ()


def stochastic_round(x, precision=0.5):
    """Performs stochastic rounding to the first decimal point."""
    scale = 1.0 / precision
    scale_x = x * scale
    fraction = scale_x - tf.floor(scale_x)

    result = tf.where(fraction < tf.random.uniform(tf.shape(x)), tf.math.floor(scale_x), tf.math.ceil(scale_x))
    return result / scale


def _round_through(x, use_stochastic_rounding=False, precision=0.5, training=True):
    """Rounds x but using straight through estimator.

    We use the trick from [Sergey Ioffe](http://stackoverflow.com/a/36480182).

    Straight through estimator is a biased estimator for the rounding
    operation defined by Hinton"s Coursera Lecture 9c where dL/dx is made
    equal to dL/dy for y = f(x) during gradient computation, where f(x) is
    a non-derivable function. In that case, we assume df/dx = 1 in:

    dL   dL df   dL
    -- = -- -- = --
    dx   df dx   dy

    (https://www.youtube.com/watch?v=LN0xtUuJsEI&list=PLoRl3Ht4JOcdU872GhiYWf6jwrk_SNhz9&index=41)

    Arguments:
      x: tensor to perform round operation with straight through gradient.
      use_stochastic_rounding: if true, we perform stochastic rounding.
      precision: by default we will use 0.5 as precision, but that can overriden
        by the user.

    Returns:
      Rounded tensor.
    """
    if use_stochastic_rounding:
        output = tf.cond(
            tf.cast(training, tf.bool),
            lambda: tf.add(x, tf.stop_gradient(-x + stochastic_round(x, precision))),
            lambda: tf.add(x, tf.stop_gradient(-x + tf.round(x))),
        )
    else:
        output = tf.add(x, tf.stop_gradient(-x + tf.round(x)))

    return tf.convert_to_tensor(output)


def _get_scaling_axis(scale_axis, len_axis):
    """Get the axis to perform auto scaling with."""

    if scale_axis is not None:
        axis = list(range(scale_axis))
        axis += list(range(scale_axis + 1, len_axis))
    else:
        if keras.backend.image_data_format() == 'channels_last':
            axis = list(range(len_axis - 1))
        else:
            axis = list(range(1, len_axis))
    return axis


def _get_scale(alpha, x, q, scale_axis=None, per_channel_scale=True):
    """Gets scaling factor for scaling the tensor per channel.
    It uses the least squares method to find the scaling factor.

    (https://en.wikipedia.org/wiki/Linear_least_squares)

    Arguments:
      alpha: A float or string. When it is string, it should be either "auto" or
        "auto_po2", and scale = sum(x * q, axis=all but last) / sum(q * q,
        axis=all but last)
       x: A tensor object. Its elements are in float.
       q: A tensor object. Its elements are in quantized format of x.
       scale_axis: which axis to calculate scale from
       per_channel_scale: A bool. Whether to perform per-channel scaling or not.

    Returns:
      A scaling factor tensor or scalar for scaling tensor per channel.
    """

    if isinstance(alpha, str) and 'auto' in alpha:
        assert alpha in ['auto', 'auto_po2']
        # in different tensorflow version (e.g., 2.4)
        # x.shape is a tuple which doesn't have as_list() method
        try:
            x_shape = x.shape.as_list()
        except AttributeError:
            x_shape = list(x.shape)

        len_axis = len(x_shape)
        if not per_channel_scale:
            qx = ops.mean(x * q, keepdims=True)
            qq = ops.mean(q * q, keepdims=True)
        else:
            if len_axis > 1:
                axis = _get_scaling_axis(scale_axis, len_axis)
                qx = ops.mean(tf.math.multiply(x, q), axis=axis, keepdims=True)
                qq = ops.mean(tf.math.multiply(q, q), axis=axis, keepdims=True)
            else:
                # No summing (averaging) along the channel axis to get per-channel
                # scales.
                qx = x * q
                qq = q * q

        scale = qx / (qq + keras.backend.epsilon())
        if alpha == 'auto_po2':
            scale = ops.power(2.0, tf.math.round(ops.log(scale + keras.backend.epsilon()) / np.log(2.0)))
    elif alpha is None:
        scale = 1.0
    elif isinstance(alpha, np.ndarray):
        scale = alpha
    else:
        scale = float(alpha)
    return scale


_default_sigmoid_type = 'hard'
_sigmoid = None


def hard_sigmoid(x):
    """Computes hard_sigmoid function that saturates between 0 and 1."""

    return ops.clip(0.5 * x + 0.5, 0.0, 1.0)


def smooth_sigmoid(x):
    """Implements a linear approximation of a sigmoid function."""

    # if we use 2.65 as the clipping point, MSE w.r.t. original sigmoid is
    # smaller than hard_simoid but the arithmetic for it is (x >> 3) +
    # (x >> 4) + 0.5, which is also not bad.

    return ops.clip(0.1875 * x + 0.5, 0.0, 1.0)


def set_internal_sigmoid(mode):
    """Sets _sigmoid to either real, hard or smooth."""

    global _sigmoid

    if mode not in ['real', 'hard', 'smooth']:
        raise ValueError("mode has to be 'real', 'hard' or 'smooth'.")

    if mode == 'hard':
        _sigmoid = hard_sigmoid
    elif mode == 'smooth':
        _sigmoid = smooth_sigmoid
    elif mode == 'real':
        _sigmoid = ops.sigmoid


set_internal_sigmoid(_default_sigmoid_type)


class quantized_relu(BaseQuantizer):  # pylint: disable=invalid-name
    """Computes a quantized relu to a number of bits.

    Modified from:

    [https://github.com/BertMoons/QuantizedNeuralNetworks-Keras-Tensorflow]

    Assume h(x) = +1 with p = sigmoid(x), -1 otherwise, the expected value of
    h(x) is:

    E[h(x)] = +1 P(p <= sigmoid(x)) - 1 P(p > sigmoid(x))
            = +1 P(p <= sigmoid(x)) - 1 ( 1 - P(p <= sigmoid(x)) )
            = 2 P(p <= sigmoid(x)) - 1
            = 2 sigmoid(x) - 1, if p is sampled from a uniform distribution U[0,1]

    If use_sigmoid is 0, we just keep the positive numbers up to
    2**integer * (1 - 2**(-bits)) instead of normalizing them, which is easier
    to implement in hardware.

    Attributes:
      bits: number of bits to perform quantization.
      integer: number of bits to the left of the decimal point.
      use_sigmoid: if true, we apply sigmoid to input to normalize it.
      negative_slope: slope when activation < 0, needs to be power of 2.
      use_stochastic_rounding: if true, we perform stochastic rounding.
      relu_upper_bound: A float representing an upper bound of the unquantized
        relu. If None, we apply relu without the upper bound when
        "is_quantized_clip" is set to false (true by default).
        Note: The quantized relu uses the quantization parameters (bits and
        integer) to upper bound. So it is important to set relu_upper_bound
        appropriately to the quantization parameters. "is_quantized_clip"
        has precedence over "relu_upper_bound" for backward compatibility.
      is_quantized_clip: A boolean representing whether the inputs are clipped to
        the maximum value represented by the quantization parameters. This
        parameter is deprecated, and the default is set to True for backwards
        compatibility. Users are encouraged to use "relu_upper_bound" instead.
      qnoise_factor: float. a scalar from 0 to 1 that represents the level of
        quantization noise to add. This controls the amount of the quantization
        noise to add to the outputs by changing the weighted sum of
        (1 - qnoise_factor)*unquantized_x + qnoise_factor*quantized_x.
      var_name: String or None. A variable name shared between the tf.Variables
        created in the build function. If None, it is generated automatically.
      use_ste: Bool. Whether to use "straight-through estimator" (STE) method or
          not.
      use_variables: Bool. Whether to make the quantizer variables to be dynamic
        tf.Variables or not.

    Returns:
      Function that performs relu + quantization to bits >= 0.
    """

    def __init__(
        self,
        bits=8,
        integer=0,
        use_sigmoid=0,
        negative_slope=0.0,
        use_stochastic_rounding=False,
        relu_upper_bound=None,
        is_quantized_clip=True,
        qnoise_factor=1.0,
        var_name=None,
        use_ste=True,
        use_variables=False,
    ):
        super().__init__()
        self.bits = bits
        self.integer = integer
        self.use_sigmoid = use_sigmoid
        self.negative_slope = negative_slope
        self.use_stochastic_rounding = use_stochastic_rounding
        self.relu_upper_bound = relu_upper_bound
        self.is_quantized_clip = is_quantized_clip
        self.qnoise_factor = qnoise_factor
        self.use_ste = use_ste
        assert negative_slope >= 0.0
        if negative_slope != 0.0:
            assert np.mod(np.log2(negative_slope), 1) == 0
        self.var_name = var_name
        self.use_variables = use_variables

    def __str__(self):
        # Converts Tensors to printable strings by converting to a numpy array and
        # then using regex to remove brackets when there is only one integer bit
        integer_bits = re.sub(
            r'\[(\d)\]', r'\g<1>', str(self.integer.numpy() if isinstance(self.integer, tf.Variable) else self.integer)
        )

        flags = [str(self.bits), integer_bits]
        if self.use_sigmoid or self.use_stochastic_rounding:
            flags.append(str(int(self.use_sigmoid)))
        if self.negative_slope:
            flags.append(str(self.negative_slope))
        if self.use_stochastic_rounding:
            flags.append(str(int(self.use_stochastic_rounding)))
        return 'quantized_relu(' + ','.join(flags) + ')'

    def __call__(self, x, training=True):
        if not self.built:
            self.build(var_name=self.var_name, use_variables=self.use_variables)

        non_sign_bits = self.bits - (self.negative_slope != 0.0)
        x = ops.cast(x, dtype='float32')
        m = ops.cast(ops.power(2, non_sign_bits), dtype='float32')
        m_i = ops.cast(ops.power(2, self.integer), dtype='float32')

        # is_quantized_clip has precedence over relu_upper_bound for backward
        # compatibility.
        m_f = ops.cast(
            ops.power(tf.constant(2.0, tf.float32), ops.cast(self.integer, dtype='float32') - non_sign_bits), dtype='float32'
        )
        if self.is_quantized_clip:
            x_u = tf.where(x <= m_i - m_f, ops.relu(x), tf.ones_like(x) * (m_i - m_f))
        elif self.relu_upper_bound is not None:
            x_u = tf.where(x <= self.relu_upper_bound, ops.relu(x), tf.ones_like(x) * self.relu_upper_bound)
        else:
            x_u = ops.relu(x)

        if self.use_sigmoid:
            p = _sigmoid(x / m_i) * m
            xq = m_i * keras.backend.clip(
                2.0 * (_round_through(p, self.use_stochastic_rounding, training=training) / m) - 1.0, 0.0, 1.0 - 1.0 / m
            )
            if self.negative_slope > 0:
                neg_factor = 1 / (self.negative_slope * m)
                xq = xq + m_i * self.negative_slope * keras.backend.clip(
                    2.0 * (_round_through(p * self.negative_slope, self.use_stochastic_rounding, training=training) * neg_factor)
                    - 1.0,
                    -1.0,
                    0.0,
                )
        else:
            p = x * m / m_i
            xq = m_i * ops.clip(_round_through(p, self.use_stochastic_rounding, training=training) / m, 0.0, 1.0 - 1.0 / m)
            if self.negative_slope > 0:
                neg_factor = 1 / (self.negative_slope * m)
                xq = xq + m_i * self.negative_slope * (
                    keras.backend.clip(
                        _round_through(p * self.negative_slope, self.use_stochastic_rounding, training=training) * neg_factor,
                        -1.0,
                        0.0,
                    )
                )

        if self.relu_upper_bound and not self.is_quantized_clip:
            xq = tf.where(xq <= self.relu_upper_bound, xq, tf.ones_like(xq) * self.relu_upper_bound)

        if self.use_ste:
            return x_u + tf.stop_gradient(self.qnoise_factor * (-x_u + xq))
        else:
            return (1 - self.qnoise_factor) * x_u + tf.stop_gradient(self.qnoise_factor * xq)

    def max(self):
        """Get the maximum value that quantized_relu can represent."""
        unsigned_bits = self.bits - (self.negative_slope != 0.0)

        if unsigned_bits > 0:
            return max(1.0, np.array(ops.power(2.0, ops.cast(self.integer, dtype='float32')), dtype='float32'))
        else:
            return 1.0

    def min(self):
        """Get the minimum value that quantized_relu can represent."""
        if self.negative_slope == 0.0:
            return 0.0

        unsigned_bits = self.bits - 1
        if unsigned_bits > 0:
            return min(
                -0.0, -self.negative_slope * np.array(ops.power(2.0, ops.cast(self.integer, dtype='float32')), dtype='float32')
            )
        else:
            return -1.0

    def range(self):
        """Returns a list of all values that quantized_relu can represent

        ordered by their binary representation ascending.
        """
        assert self.use_sigmoid == 0  # current unsupported
        assert self.negative_slope == 0  # # unsupported unsupported
        x = np.asarray(range(2**self.bits))
        return x * np.array(ops.power(2.0, -self.bits + ops.cast(self.integer, dtype='float32')), dtype='float32')

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        config = {
            'bits': self.bits,
            'integer': self.integer.numpy() if isinstance(self.integer, tf.Variable) else self.integer,
            'use_sigmoid': self.use_sigmoid,
            'negative_slope': self.negative_slope,
            'use_stochastic_rounding': self.use_stochastic_rounding,
            'relu_upper_bound': self.relu_upper_bound,
            'qnoise_factor': self.qnoise_factor.numpy() if isinstance(self.qnoise_factor, tf.Variable) else self.qnoise_factor,
        }
        return config


class bernoulli(BaseQuantizer):  # pylint: disable=invalid-name
    """Computes a Bernoulli sample with probability sigmoid(x).

    This computation uses ST approximation.

    To do that, we compute sigmoid(x) and a random sample z ~ U[0,1]. As
    p in [0,1] and z in [0,1], p - z in [-1,1]. However, -1 will
    never appear because to get -1 we would need sigmoid(-inf) - z == 1.
    As a result, the range will be in practical terms [0,1].

    The noise introduced by z can be seen as a regularizer to the weights W of
    y = Wx as y = Wx + Wz for some noise z with mean mu(z) and var(z). As a
    result, W**2 var(z) to the variance of y, which has the same effect as a
    regularizer on L2 with lambda = var(z), as presented in Hinton"s Coursera
    Lecture 9c.

    Remember that E[dL/dy] = E[dL/dx] once we add stochastic sampling.

    Attributes:
      alpha: allows one to specify multiplicative factor for number generation
        of "auto" or "auto_po2".
      temperature: amplifier factor for sigmoid function, making stochastic
        less stochastic as it moves away from 0.
      use_real_sigmoid: use real sigmoid for probability.

    Returns:
      Computation of round with stochastic sampling with straight through
      gradient.
    """

    def __init__(self, alpha=None, temperature=6.0, use_real_sigmoid=True, thr=None):
        super().__init__()
        self.alpha = alpha
        self.bits = 1
        self.temperature = temperature
        self.use_real_sigmoid = use_real_sigmoid
        self.default_alpha = 1.0
        self.scale = None
        if thr is not None:
            self.thr = tf.Variable(tf.ones([1]) * thr, trainable=True, dtype=tf.float32)
        else:
            self.thr = None

    def __str__(self):
        flags = []
        if self.alpha is not None:
            alpha = str(self.alpha)
            if isinstance(self.alpha, str):
                alpha = "'" + alpha + "'"
            flags.append('alpha=' + alpha)
        if self.temperature != 6.0:
            flags.append('temperature=' + str(self.temperature))
        if not self.use_real_sigmoid:
            flags.append('use_real_sigmoid=' + str(int(self.use_real_sigmoid)))
        return 'bernoulli(' + ','.join(flags) + ')'

    def __call__(self, x, training=True):
        if isinstance(self.alpha, str):
            assert self.alpha in ['auto', 'auto_po2']

        if isinstance(self.alpha, str):
            len_axis = len(x.shape)

            if len_axis > 1:
                if ops.image_data_format() == 'channels_last':
                    axis = list(range(len_axis - 1))
                else:
                    axis = list(range(1, len_axis))
            else:
                axis = [0]

            std = ops.std(x, axis=axis, keepdims=True) + keras.backend.epsilon()
        else:
            std = 1.0

        if self.use_real_sigmoid:
            p = keras.backend.sigmoid(self.temperature * x / std)
        else:
            p = _sigmoid(self.temperature * x / std)

        if training or self.thr is None:
            r = tf.random.uniform(tf.shape(x))
            q = tf.sign(p - r)
            q += 1.0 - tf.abs(q)
            q = (q + 1.0) / 2.0
        else:
            q = tf.where(p >= self.thr, tf.ones_like(p), tf.zeros_like(p))

        q_non_stochastic = tf.sign(x)
        q_non_stochastic += 1.0 - tf.abs(q_non_stochastic)
        q_non_stochastic = (q_non_stochastic + 1.0) / 2.0

        # if we use non stochastic binary to compute alpha,
        # this function seems to behave better
        scale = _get_scale(self.alpha, x, q_non_stochastic)
        self.scale = scale
        return x + tf.stop_gradient(-x + scale * q)

    def _set_trainable_parameter(self):
        if self.alpha is None:
            self.alpha = 'auto_po2'

    def max(self):
        """Get the maximum value bernoulli class can represent."""
        if self.alpha is None or isinstance(self.alpha, str):
            return 1.0
        else:
            return max(1.0, self.alpha)

    def min(self):
        """Get the minimum value bernoulli class can represent."""
        return 0.0

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        config = {'alpha': self.alpha, 'thr': self.thr.numpy().tolist() if self.thr is not None else None}
        return config



class BernoulliSampling(keras.layers.Layer):
    def __init__(self, num_samples=1, name=None, std=1, thr=0.5, temperature=6.0, use_quantized=False, bits_bernoulli_sigmoid=8, **kwargs):
        super().__init__(name=name, **kwargs)
        self.num_samples = num_samples
        self.std = std
        self.temperature = temperature
        self.use_quantized = use_quantized
        self.bits_bernoulli_sigmoid = bits_bernoulli_sigmoid
        if thr is not None:
            self.thr = tf.Variable(tf.ones([1]) * thr, trainable=False, dtype=tf.float32)
        else:
            self.thr = None

    def call(self, inputs, training=True):
        # convert inputs to sigmoid to get probablity for bernoulli
        if self.use_quantized:
            p = quantized_sigmoid(bits=self.bits_bernoulli_sigmoid, use_stochastic_rounding=True, symmetric=True)(
                self.temperature * inputs / self.std
            )
            p = tf.cast(p, tf.float32)
        else:
            p = ops.sigmoid(self.temperature * inputs / self.std)

        # sample num_samples times from a bernoulli
        out = tf.zeros(tf.shape(inputs))
        if training:
            for _ in range(self.num_samples):
                r = tf.random.uniform(tf.shape(inputs))
                q = tf.sign(p - r)
                q += 1.0 - tf.abs(q)
                q = (q + 1.0) / 2.0
                out += q
            out = out / self.num_samples
        else:
            out = tf.where(p >= self.thr, tf.ones_like(p), tf.zeros_like(p))

        # output is mean of stochastic sampling with straight through gradient
        out = inputs + tf.stop_gradient(-inputs + out)

        return out

    def max(self):
        """Get the maximum value bernoulli class can represent."""
        return 1.0

    def min(self):
        """Get the minimum value bernoulli class can represent."""
        return 0.0

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        config = {'num_samples': self.num_samples, 'thr': self.thr.numpy().tolist() if self.thr is not None else None}
        return config

class QActivation(keras.layers.Layer):
    """Implements quantized activation layers."""

    def __init__(self, activation, **kwargs):
        super().__init__(**kwargs)

        self.activation = activation

        if not isinstance(activation, str):
            self.quantizer = activation
            if hasattr(self.quantizer, '__name__'):
                self.__name__ = self.quantizer.__name__
            elif hasattr(self.quantizer, 'name'):
                self.__name__ = self.quantizer.name
            elif hasattr(self.quantizer, '__class__'):
                self.__name__ = self.quantizer.__class__.__name__
            return

        self.__name__ = activation

        try:
            self.quantizer = get_quantizer(activation)
        except KeyError:
            raise ValueError(f"invalid activation '{activation}'")

    def call(self, inputs):
        return self.quantizer(inputs)

    def get_config(self):
        config = {'activation': self.activation}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def get_quantization_config(self):
        return str(self.activation)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_prunable_weights(self):
        return []


class quantized_bits(BaseQuantizer):  # pylint: disable=invalid-name
    """Quantizes the number to a number of bits.

    In general, we want to use a quantization function like:

    a = (power(2,bits) - 1 - 0) / (max(x) - min(x))
    b = -min(x) * a

    in the equation:

    xq = a x + b

    This requires multiplication, which is undesirable. So, we
    enforce weights to be between -1 and 1 (max(x) = 1 and min(x) = -1),
    and separating the sign from the rest of the number as we make this function
    symmetric, thus resulting in the following approximation.

    1) max(x) = +1, min(x) = -1
    2) max(x) = -min(x)

    a = power(2,bits-1)
    b = 0

    Finally, just remember that to represent the number with sign, the
    largest representation is -power(2,bits) to power(2, bits-1)

    Symmetric and keep_negative allow us to generate numbers that are symmetric
    (same number of negative and positive representations), and numbers that
    are positive.

    Note:
      the behavior of quantized_bits is different than Catapult HLS ac_fixed
      or Vivado HLS ap_fixed. For ac_fixed<word_length, integer_lenth, signed>,
      when signed = true, it is equavlent to
      quantized_bits(word_length, integer_length-1, keep_negative=True)

    Attributes:
      bits: number of bits to perform quantization.
      integer: number of bits to the left of the decimal point.
      symmetric: if true, we will have the same number of values for positive
        and negative numbers.
      alpha: a tensor or None, the scaling factor per channel.
        If None, the scaling factor is 1 for all channels.
      keep_negative: if true, we do not clip negative numbers.
      use_stochastic_rounding: if true, we perform stochastic rounding.
      scale_axis: which axis to calculate scale from
      qnoise_factor: float. a scalar from 0 to 1 that represents the level of
        quantization noise to add. This controls the amount of the quantization
        noise to add to the outputs by changing the weighted sum of
        (1 - qnoise_factor)*unquantized_x + qnoise_factor*quantized_x.
      var_name: String or None. A variable name shared between the tf.Variables
        created in the build function. If None, it is generated automatically.
      use_ste: Bool. Whether to use "straight-through estimator" (STE) method or
          not.
      use_variables: Bool. Whether to make the quantizer variables to be dynamic
        tf.Variables or not.

    Returns:
      Function that computes fixed-point quantization with bits.
    """

    def __init__(
        self,
        bits=8,
        integer=0,
        symmetric=0,
        keep_negative=True,
        alpha=None,
        use_stochastic_rounding=False,
        scale_axis=None,
        qnoise_factor=1.0,
        var_name=None,
        use_ste=True,
        use_variables=False,
    ):
        super().__init__()
        self.bits = bits
        self.integer = integer
        self.symmetric = symmetric
        self.keep_negative = keep_negative
        self.alpha = alpha
        self.use_stochastic_rounding = use_stochastic_rounding
        # "auto*" |-> symmetric
        if isinstance(self.alpha, str):
            self.symmetric = True
        self.scale = None
        self.scale_axis = scale_axis
        self.qnoise_factor = qnoise_factor
        self.use_ste = use_ste
        self.var_name = var_name
        self.use_variables = use_variables

    def __str__(self):
        # Convert Tensors to printable strings by converting to a numpy array and
        # then using regex to remove brackets when there is only one integer bit
        integer_bits = re.sub(
            r'\[(\d)\]', r'\g<1>', str(self.integer.numpy() if isinstance(self.integer, tf.Variable) else self.integer)
        )

        flags = [str(self.bits), integer_bits, str(int(self.symmetric))]
        if not self.keep_negative:
            flags.append('keep_negative=False')
        if self.alpha:
            alpha = str(self.alpha)
            if isinstance(self.alpha, str):
                alpha = "'" + alpha + "'"
            flags.append('alpha=' + alpha)
        if self.use_stochastic_rounding:
            flags.append('use_stochastic_rounding=' + str(int(self.use_stochastic_rounding)))
        return 'quantized_bits(' + ','.join(flags) + ')'

    def __call__(self, x, training=True):
        """Computes fixedpoint quantization of x."""
        if not self.built:
            self.build(var_name=self.var_name, use_variables=self.use_variables)

        x = ops.cast(tf.convert_to_tensor(x), dtype='float32')

        # quantized_bits with "1" bit becomes a binary implementation.
        unsigned_bits = self.bits - self.keep_negative
        m = ops.cast(ops.power(2, unsigned_bits), dtype='float32')
        m_i = ops.cast(ops.power(2, self.integer), dtype='float32')

        if self.alpha is None:
            scale = 1.0
        elif isinstance(self.alpha, str):
            # We only deal with the symmetric case right now.
            assert self.symmetric, 'Only symmetric quantizers are implemented'
            len_axis = len(x.shape)
            if len_axis > 1:
                axis = _get_scaling_axis(self.scale_axis, len_axis)
            else:
                axis = [0]

            x = x / m_i

            # Using 2's complement, we can represent 2**(bits-1)-1 positive values
            # If we wish to maintain symmetry, we can double 2**(bits-1)-1 to get
            # the total number of possible values we can represent.
            # If symmetry is not enforced, then we can represent (2**bits)-1 values
            # using 2's complement.
            levels = (2 ** (self.bits - 1) - 1) * 2 if self.symmetric else (2**self.bits) - 1

            scale = (ops.max(abs(x), axis=axis, keepdims=True) * 2) / levels

            # If alpha is "auto_po2", then get the "best" po2 scale
            if 'po2' in self.alpha:
                scale = ops.power(2.0, tf.math.round(ops.log(scale + keras.backend.epsilon()) / np.log(2.0)))
                for _ in range(5):
                    v = tf.floor(tf.abs(x) / scale + 0.5)
                    mask = v < levels / 2
                    z = tf.sign(x) * tf.where(mask, v, tf.ones_like(v) * levels / 2)
                    scale = _get_scale(alpha='auto_po2', x=x, q=z, scale_axis=self.scale_axis)

            # If alpha is "auto", then get the "best" floating point scale
            elif self.alpha == 'auto':
                v = tf.floor(tf.abs(x) / scale + 0.5)
                mask = v < levels / 2
                z = tf.sign(x) * tf.where(mask, v, tf.ones_like(v) * levels / 2)
            else:
                raise ValueError(f"Invalid alpha '{self.alpha}'")

            # z is an integer number, so we must make the scale * m and z / m
            scale = scale * m

            # we will not use "z" right now because of stochastic_rounding
            # this is still under test.

            # if "new" in self.alpha:
            #  z = z / m
            #  self.scale = scale
            #  return x + tf.stop_gradient(-x + scale * z)
            x = m_i * x
            xq = m_i * z / m
            self.scale = scale
            xq = scale * xq

            if self.use_ste:
                return x + tf.stop_gradient(self.qnoise_factor * (-x + xq))
            else:
                return (1 - self.qnoise_factor) * x + tf.stop_gradient(self.qnoise_factor * xq)

        else:
            scale = self.alpha

        def handle_unsigned_bits(x, m, m_i, unsigned_bits):
            p = x * m / m_i
            xq = (
                m_i
                * tf.clip_by_value(
                    _round_through(p, self.use_stochastic_rounding, precision=1.0, training=training),
                    self.keep_negative * (-m + self.symmetric),
                    m - 1,
                )
                / m
            )
            return xq

        def handle_binary_quantization(x, keep_negative):
            xq = tf.sign(x)
            xq += 1.0 - tf.abs(xq)
            if not keep_negative:
                xq = (xq + 1.0) / 2.0
            return xq

        # quantized_bits with "1" bit becomes a binary implementation.
        xq = tf.cond(
            tf.cast(unsigned_bits > 0, tf.bool),  # Condition
            lambda: handle_unsigned_bits(x, m, m_i, unsigned_bits),  # True case
            lambda: handle_binary_quantization(x, self.keep_negative),  # False case
        )

        self.scale = scale
        xq = scale * xq

        if self.use_ste:
            return x + tf.stop_gradient(self.qnoise_factor * (-x + xq))
        else:
            return (1 - self.qnoise_factor) * x + tf.stop_gradient(self.qnoise_factor * xq)

    def _set_trainable_parameter(self):
        if self.alpha is None:
            self.alpha = 'auto_po2'
            self.symmetric = True

    def max(self):
        """Get maximum value that quantized_bits class can represent."""
        unsigned_bits = self.bits - self.keep_negative
        if unsigned_bits > 0:
            return max(1.0, np.array(ops.power(2.0, ops.cast(self.integer, dtype='float32')), dtype='float32'))
        else:
            return 1.0

    def min(self):
        """Get minimum value that quantized_bits class can represent."""
        if not self.keep_negative:
            return 0.0
        unsigned_bits = self.bits - self.keep_negative
        if unsigned_bits > 0:
            return -max(1.0, np.array(keras.ops.power(2, keras.ops.cast(self.integer, dtype='float32')), dtype='float32'))
        else:
            return -1.0

    def range(self):
        """Returns a list of all values that quantized_bits can represent
        ordered by their binary representation ascending."""
        assert self.symmetric == 0
        assert self.keep_negative
        assert self.alpha is None or self.alpha == 1.0

        x = np.asarray(range(2**self.bits), dtype=np.float32)
        p_and_n = np.where(x >= 2 ** (self.bits - 1), (x - 2 ** (self.bits - 1)) - 2 ** (self.bits - 1), x)
        return p_and_n * np.array(
            keras.ops.power(2.0, -self.bits + keras.ops.cast(self.integer, dtype='float32') + 1), dtype='float32'
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        config = {
            'bits': self.bits,
            'integer': self.integer.numpy() if isinstance(self.integer, tf.Variable) else self.integer,
            'symmetric': self.symmetric,
            'alpha': self.alpha,
            'keep_negative': self.keep_negative,
            'use_stochastic_rounding': self.use_stochastic_rounding,
            'qnoise_factor': self.qnoise_factor.numpy() if isinstance(self.qnoise_factor, tf.Variable) else self.qnoise_factor,
        }
        return config


class Clip(keras.constraints.Constraint):
    """Clips weight constraint."""

    # This function was modified from Keras minmaxconstraints.
    #
    # Constrains the weights to be between min/max values.
    #   min_value: the minimum norm for the incoming weights.
    #   max_value: the maximum norm for the incoming weights.
    #   constraint: previous constraint to be clipped.
    #   quantizer: quantizer to be applied to constraint.

    def __init__(self, min_value=0.0, max_value=1.0, constraint=None, quantizer=None):
        """Initializes Clip constraint class."""

        self.min_value = min_value
        self.max_value = max_value
        self.constraint = keras.constraints.get(constraint)
        # Don't wrap yourself
        if isinstance(self.constraint, Clip):
            self.constraint = None
        self.quantizer = get_quantizer(quantizer)

    def __call__(self, w):
        """Clips values between min and max values."""
        if self.constraint:
            w = self.constraint(w)
            if self.quantizer:
                w = self.quantizer(w)
        w = ops.clip(w, self.min_value, self.max_value)
        return w

    def get_config(self):
        """Returns configuration of constraint class."""
        return {'min_value': self.min_value, 'max_value': self.max_value}

    @classmethod
    def from_config(cls, config):
        if isinstance(config.get('constraint', None), Clip):
            config['constraint'] = None
        config['constraint'] = keras.constraints.get(config.get('constraint', None))
        config['quantizer'] = get_quantizer(config.get('quantizer', None))
        return cls(**config)


def get_initializer(identifier):
    """Gets the initializer.

    Args:
      identifier: An initializer, which could be dict, string, or callable function.

    Returns:
      A initializer class

    Raises:
      ValueError: An error occurred when quantizer cannot be interpreted.
    """
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        if identifier['class_name'] == 'QInitializer':
            return QInitializer.from_config(identifier['config'])
        else:
            return keras.initializers.get(identifier)
    elif isinstance(identifier, str):
        return keras.initializers.get(identifier)
    elif callable(identifier):
        return identifier
    else:
        raise ValueError('Could not interpret initializer identifier: ' + str(identifier))


class QInitializer(keras.initializers.Initializer):
    """Wraps around Keras initializer to provide a fanin scaling factor."""

    def __init__(self, initializer, use_scale, quantizer):
        self.initializer = initializer
        self.use_scale = use_scale
        self.quantizer = quantizer

        try:
            self.is_po2 = 'po2' in quantizer.__class__.__name__
        except:
            self.is_po2 = False

    def __call__(self, shape, dtype=None):
        x = self.initializer(shape, dtype)

        std_x = np.std(x)
        delta = self.quantizer.max() * 2**-self.quantizer.bits

        # delta is the minimum resolution of the number system.
        # we want to make sure we have enough values.
        if delta > std_x and hasattr(self.initializer, 'scale'):
            q = self.quantizer(x)
            max_q = np.max(abs(q))
            scale = 1.0
            if max_q == 0.0:
                xx = np.mean(x * x)
                scale = self.quantizer.max() / np.sqrt(xx)
            else:
                qx = np.sum(q * x)
                qq = np.sum(q * q)

                scale = qq / qx

            self.initializer.scale *= max(scale, 1)
            x = self.initializer(shape, dtype)

        return np.clip(x, -self.quantizer.max(), self.quantizer.max())

    def get_config(self):
        return {
            'initializer': self.initializer,
            'use_scale': self.use_scale,
            'quantizer': self.quantizer,
        }

    @classmethod
    def from_config(cls, config):
        config = {
            'initializer': get_initializer(config['initializer']),
            'use_scale': config['use_scale'],
            'quantizer': get_quantizer(config['quantizer']),
        }
        return cls(**config)


def get_constraint(identifier, quantizer):
    """Gets the initializer.

    Args:
      identifier: A constraint, which could be dict, string, or callable function.
      quantizer: A quantizer class or quantization function

    Returns:
      A constraint class
    """
    if identifier:
        if isinstance(identifier, dict) and identifier['class_name'] == 'Clip':
            return Clip.from_config(identifier['config'])
        else:
            return keras.constraints.get(identifier)
    else:
        max_value = max(1, quantizer.max()) if hasattr(quantizer, 'max') else 1.0
        return Clip(-max_value, max_value, identifier, quantizer)

def _need_exponent_sign_bit_check(max_value):
  """Checks whether the sign bit of exponent is needed.

  This is used by quantized_po2 and quantized_relu_po2.

  Args:
    max_value: the maximum value allowed.

  Returns:
    An integer. 1: sign_bit is needed. 0: sign_bit is not needed.
  """

  if max_value is not None:
    if max_value < 0:
      raise ValueError("po2 max_value should be non-negative.")
    if max_value > 1:
      # if max_value is larger than 1,
      #   the exponent could be positive and negative.
      #   e.g., log(max_value) > 0 when max_value > 1
      need_exponent_sign_bit = 1
    else:
      need_exponent_sign_bit = 0
  else:
    # max_value is not specified, so we cannot decide the range.
    # Then we need to put sign_bit for exponent to be safe
    need_exponent_sign_bit = 1
  return need_exponent_sign_bit


def get_auto_range_constraint_initializer(quantizer, constraint, initializer):
    """Get value range automatically for quantizer.

    Arguments:
     quantizer: A quantizer class in quantizers.py.
     constraint: A keras constraint.
     initializer: A keras initializer.

    Returns:
      a tuple (constraint, initializer), where
        constraint is clipped by Clip class in this file, based on the
        value range of quantizer.
        initializer is initializer contraint by value range of quantizer.
    """
    if quantizer is not None:
        constraint = get_constraint(constraint, quantizer)
        initializer = get_initializer(initializer)

        if initializer and initializer.__class__.__name__ not in ['Ones', 'Zeros', 'QInitializer']:
            # we want to get the max value of the quantizer that depends
            # on the distribution and scale
            if not (hasattr(quantizer, 'alpha') and isinstance(quantizer.alpha, str)):
                initializer = QInitializer(initializer, use_scale=True, quantizer=quantizer)
    return constraint, initializer


def Num(s):
    """Tries to convert string to either int or float."""
    try:
        try:
            return int(s)
        except ValueError:
            return float(s)
    except ValueError:
        # this should be always true. if it isn't int or float, it should be str
        assert (s[0] == '"' and s[-1] == '"') or (s[0] == "'" and s[-1] == "'")
        s = s[1:-1]
        return s


def Str(s):
    return s[1:-1]


def IsNum(s):
    try:
        try:
            int(s)
            return True
        except ValueError:
            float(s)
            return True
    except ValueError:
        return False


def IsBool(s):
    if s in ['True', 'False']:
        return True
    else:
        return False


def Bool(s):
    return True if 'True' in s else False


def GetArg(s):
    if IsBool(s):
        return Bool(s)
    elif IsNum(s):
        return Num(s)
    else:
        return Str(s)


def GetParams(s):
    """Extracts args and kwargs from string."""
    # modified from https://stackoverflow.com/questions/38799223/parse-string-to-identify-kwargs-and-args  # pylint: disable=line-too-long

    _lparen = Suppress('(')  # pylint: disable=invalid-name
    _rparen = Suppress(')')  # pylint: disable=invalid-name
    _eq = Suppress('=')  # pylint: disable=invalid-name

    data = _lparen + Optional(delimitedList(Group(Regex(r'[^=,)\s]+') + Optional(_eq + Regex('[^,)]*'))))) + _rparen

    items = data.parseString(s).asList()

    # need to make sure that kwargs only happen after args are processed
    args = [GetArg(i[0]) for i in items if len(i) == 1]
    kwargs = {i[0]: GetArg(i[1]) for i in items if len(i) == 2}

    # check for syntax error
    for i in range(1, len(items)):
        if (len(items[i]) == 1) and (len(items[i - 1]) == 2):
            raise SyntaxError

    return args, kwargs


def safe_eval(eval_str, op_dict, *params, **kwparams):  # pylint: disable=invalid-name
    """Replaces eval by a safe eval mechanism."""

    function_split = eval_str.split('(')
    quantizer = op_dict.get(function_split[0], None)

    if len(function_split) == 2:
        args, kwargs = GetParams('(' + function_split[1])
    else:
        args = []
        kwargs = {}

    args = args + list(params)
    for k in kwparams:
        kwargs[k] = kwparams[k]

    # must be Keras activation object if None
    if quantizer is None:
        logging.info('keras dict %s', function_split[0])
        quantizer = keras.activations.get(function_split[0])

    if len(function_split) == 2 or args or kwargs:
        return quantizer(*args, **kwargs)
    else:
        if isinstance(quantizer, type):
            # Check if quantizer is a class
            return quantizer()
        else:
            # Otherwise it is a function, so just return it
            return quantizer


def _floor_through(x):
  """Computes the floor operation using straight through estimator."""

  return x + tf.stop_gradient(-x + tf.floor(x))


def stochastic_round_po2(x):
  """Performs stochastic rounding for the power of two."""
  # TODO(b/237832905): test stochastic_round_po2 and constraint.
  # because quantizer is applied after constraint.
  y = tf.abs(x)
  eps = tf.keras.backend.epsilon()
  log2 = tf.keras.backend.log(2.0)

  x_log2 = tf.round(tf.keras.backend.log(y + eps) / log2)
  po2 = tf.cast(pow(2.0, tf.cast(x_log2, dtype="float32")), dtype="float32")
  left_val = tf.where(po2 > y, x_log2 - 1, x_log2)
  right_val = tf.where(po2 > y, x_log2, x_log2 + 1)
  # sampling in [2**left_val, 2**right_val].
  minval = 2 ** left_val
  maxval = 2 ** right_val
  val = tf.random.uniform(tf.shape(y), minval=minval, maxval=maxval)
  # use y as a threshold to keep the probabliy [2**left_val, y, 2**right_val]
  # so that the mean value of the sample should be y
  x_po2 = tf.where(y < val, left_val, right_val)
  """
  x_log2 = stochastic_round(tf.keras.backend.log(y + eps) / log2)
  sign = tf.sign(x)
  po2 = (
      tf.sign(x) *
      tf.cast(pow(2.0, tf.cast(x_log2, dtype="float32")), dtype="float32")
  )
  """
  return x_po2


def _clip_power_of_two(x_abs,
                       min_exp,
                       max_exp,
                       max_value,
                       quadratic_approximation=False,
                       use_stochastic_rounding=False,
                       log2_rounding="rnd"):
  """Clips a tensor using power-of-two quantizer.


  Args:
    x_abs: A tensor object. Its elements should be non-negative.
    min_exp: An integer representing the smallest exponent.
    max_exp: An integer representing the largest exponent.
    max_value: A float or None. If it is None, we clip the value to max_value.
    quadratic_approximation: An boolean representing whether the quadratic
      approximation is applied.
    use_stochastic_rounding: An boolean representing whether the stochastic
      rounding method is applied.
    log2_rounding: log2 rounding mode. "rnd" and "floor" currently
      supported, corresponding to tf.round and tf.floor respectively.

  Returns:
    A tensor object, the values are clipped by min_exp and max_exp.
  """

  # if quadratic_approximation is True, round to the exponent for sqrt(x),
  # so that the return value can be divided by two without remainder.
  log2 = np.log(2.0)

  # When the elements of x_abs are small than the keras epsilon,
  # we just overwrite x_abs with eps
  eps = tf.keras.backend.epsilon()
  x_filter = tf.where(x_abs < eps, eps, x_abs)
  if max_value is not None:
    # If the elements of x_filter has value larger than x_value, clip it.
    x_filter = tf.where(x_filter >= max_value,
                        tf.ones_like(x_filter) * max_value, x_filter)

  def power_of_two_clip(x_abs, min_exp, max_exp, quadratic_approximation,
                        use_stochastic_rounding, log2_rounding):
    assert log2_rounding in ["rnd", "floor"]

    if quadratic_approximation:
      q_factor = 2.0
      x_input = tf.sqrt(x_abs)
    else:
      q_factor = 1.0
      x_input = x_abs

    if log2_rounding == "floor":
      x_log2 = _floor_through(tf.keras.backend.log(x_input) / log2)
    elif use_stochastic_rounding:
      x_log2 = tf.keras.tf_utils.smart_cond(
          tf.keras.learning_phase(),
          lambda: stochastic_round_po2(x_input),
          lambda: _round_through(tf.keras.backend.log(x_input) / log2))
    else:
      x_log2 = _round_through(tf.keras.backend.log(x_input) / log2)

    x_clipped = q_factor * tf.keras.backend.clip(x_log2, min_exp, max_exp)
    return x_clipped

  x_clipped = tf.where(
      x_abs < eps,
      tf.ones_like(x_abs) * min_exp,
      power_of_two_clip(x_filter, min_exp, max_exp, quadratic_approximation,
                        use_stochastic_rounding, log2_rounding))

  return x_clipped


def _get_min_max_exponents(non_sign_bits, need_exponent_sign_bit,
                           quadratic_approximation):
  """Given a bitwidth, gets min and max exponents that it can represent.

  Args:
    non_sign_bits: An integer representing the bitwidth of the exponent.
    need_exponent_sign_bit: An integer representing whether it needs sign bit
      in exponent. (1: need sign bit. 0: sign bit is not needed.)
    quadratic_approximation: A boolean representing whether the quadratic
      approximiation method is enforced.

  Returns:
    A tuple of integers: min_exp, max_exp
  """
  effect_bits = non_sign_bits - need_exponent_sign_bit
  min_exp = -2**(effect_bits)
  max_exp = 2**(effect_bits) - 1
  if quadratic_approximation:
    max_exp = 2 * (max_exp // 2)
  return min_exp, max_exp


def get_quantizer(identifier):
    """Gets the quantizer.

    Args:
      identifier: An quantizer, which could be dict, string, or callable function.

    Returns:
      A quantizer class or quantization function from this file. For example,
        Quantizer classes: quantized_bits, quantized_po2, quantized_relu_po2,
        binary, stochastic_binary, ternary, stochastic_ternary, etc.

        Quantization functions: binary_sigmoid, hard_sigmoid, soft_sigmoid, etc.

    Raises:
      ValueError: An error occurred when quantizer cannot be interpreted.
    """

    if identifier is None:
        return None
    if isinstance(identifier, dict):
        return keras.utils.deserialize_keras_object(identifier, module_objects=globals(), printable_module_name='quantizer')
    elif isinstance(identifier, str):
        return safe_eval(identifier, globals())
    elif callable(identifier):
        return identifier
    else:
        raise ValueError('Could not interpret quantizer identifier: ' + str(identifier))


class quantized_po2(BaseQuantizer):  # pylint: disable=invalid-name
  """Quantizes to the closest power of 2.

  Attributes:
    bits: An integer, the bits allocated for the exponent, its sign and the sign
      of x.
    max_value: An float or None. If None, no max_value is specified.
      Otherwise, the maximum value of quantized_po2 <= max_value
    use_stochastic_rounding: A boolean, default is False, if True, it uses
      stochastic rounding and forces the mean of x to be x statstically.
    quadratic_approximation: A boolean, default is False if True, it forces the
      exponent to be even number that closted to x.
    log2_rounding: A string, log2 rounding mode. "rnd" and "floor" currently
      supported, corresponding to tf.round and tf.floor respectively.
    qnoise_factor: float. a scalar from 0 to 1 that represents the level of
      quantization noise to add. This controls the amount of the quantization
      noise to add to the outputs by changing the weighted sum of
      (1 - qnoise_factor)*unquantized_x + qnoise_factor*quantized_x.
    var_name: String or None. A variable name shared between the tf.Variables
      created in the build function. If None, it is generated automatically.
    use_ste: Bool. Whether to use "straight-through estimator" (STE) method or
        not.
    use_variables: Bool. Whether to make the quantizer variables to be dynamic
      tf.Variables or not.
  """

  def __init__(self,
               bits=8,
               max_value=None,
               use_stochastic_rounding=False,
               quadratic_approximation=False,
               log2_rounding="rnd",
               qnoise_factor=1.0,
               var_name=None,
               use_ste=True,
               use_variables=False):
    super().__init__()
    self.bits = bits
    self.max_value = max_value
    self.use_stochastic_rounding = use_stochastic_rounding
    self.log2_rounding = log2_rounding
    # if True, round to the exponent for sqrt(x),
    # so that the return value can be divided by two without remainder.
    self.quadratic_approximation = quadratic_approximation
    need_exponent_sign_bit = _need_exponent_sign_bit_check(self.max_value)
    non_sign_bits = self.bits - 1
    self._min_exp, self._max_exp = _get_min_max_exponents(
        non_sign_bits, need_exponent_sign_bit, self.quadratic_approximation)
    # qnoise_factor related attributes
    self.qnoise_factor = qnoise_factor
    self.use_ste = use_ste
    self.var_name = var_name
    self.use_variables = use_variables

  def __str__(self):
    flags = [str(self.bits)]
    if self.max_value is not None or self.use_stochastic_rounding:
      flags.append(str(int(self.max_value)))
    if self.use_stochastic_rounding:
      flags.append(str(int(self.use_stochastic_rounding)))
    if self.quadratic_approximation:
      flags.append(
          "quadratic_approximation=" + str(int(self.quadratic_approximation)))
    return "quantized_po2(" + ",".join(flags) + ")"

  def __call__(self, x):
    if not self.built:
      self.build(var_name=self.var_name, use_variables=self.use_variables)

    x_sign = tf.sign(x)
    x_sign += (1.0 - tf.abs(x_sign))
    x_abs = tf.abs(x)
    x_clipped = _clip_power_of_two(x_abs, self._min_exp, self._max_exp,
                                   self.max_value,
                                   self.quadratic_approximation,
                                   self.use_stochastic_rounding,
                                   self.log2_rounding)
    xq = x_sign * pow(2.0, x_clipped)

    if self.use_ste:
      return x + tf.stop_gradient(self.qnoise_factor * (-x + xq))
    else:
      return (1 - self.qnoise_factor) * x + tf.stop_gradient(
          self.qnoise_factor * xq)

  def max(self):
    """Get the maximum value that quantized_po2 can represent."""
    if self.max_value:
      return max(1.0, self.max_value)
    else:
      return max(1.0, 2**self._max_exp)

  def min(self):
    """Get the minimum value that quantized_po2 can represent."""
    if self.max_value:
      return -max(1.0, self.max_value)
    else:
      return -max(1.0, 2**self._max_exp)

  @classmethod
  def from_config(cls, config):
    return cls(**config)

  def get_config(self):
    """Gets configugration of the quantizer.

    Returns:
      A dict mapping quantization configuration, including
        bits: bitwidth for exponents.
        max_value: the maximum value of this quantized_po2 can represent.
        use_stochastic_rounding:
          if True, stochastic rounding is used.
        quadratic_approximation:
          if True, the exponent is enforced to be even number, which is
          the closest one to x.
        log2_rounding:
          A string, Log2 rounding mode
    """
    config = {
        "bits":
            self.bits,
        "max_value":
            self.max_value,
        "use_stochastic_rounding":
            self.use_stochastic_rounding,
        "quadratic_approximation":
            self.quadratic_approximation,
        "qnoise_factor":
            self.qnoise_factor.numpy() if isinstance(
                self.qnoise_factor, tf.Variable) else self.qnoise_factor,
        "log2_rounding":
            self.log2_rounding
    }
    return config


class QDense(keras.layers.Dense):
    """Implements a quantized Dense layer."""

    # Most of these parameters follow the implementation of Dense in
    # Keras, with the exception of kernel_range, bias_range,
    # kernel_quantizer, bias_quantizer, and kernel_initializer.
    #
    # kernel_quantizer: quantizer function/class for kernel
    # bias_quantizer: quantizer function/class for bias
    # kernel_range/bias_ranger: for quantizer functions whose values
    #   can go over [-1,+1], these values are used to set the clipping
    #   value of kernels and biases, respectively, instead of using the
    #   constraints specified by the user.
    #
    # we refer the reader to the documentation of Dense in Keras for the
    # other parameters.

    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer='he_normal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kernel_quantizer=None,
        bias_quantizer=None,
        kernel_range=None,
        bias_range=None,
        **kwargs,
    ):
        if kernel_range is not None:
            warnings.warn('kernel_range is deprecated in QDense layer.')

        if bias_range is not None:
            warnings.warn('bias_range is deprecated in QDense layer.')

        self.kernel_range = kernel_range
        self.bias_range = bias_range

        self.kernel_quantizer = kernel_quantizer
        self.bias_quantizer = bias_quantizer

        self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
        self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

        # optimize parameter set to "auto" scaling mode if possible
        if hasattr(self.kernel_quantizer_internal, '_set_trainable_parameter'):
            self.kernel_quantizer_internal._set_trainable_parameter()

        self.quantizers = [self.kernel_quantizer_internal, self.bias_quantizer_internal]

        kernel_constraint, kernel_initializer = get_auto_range_constraint_initializer(
            self.kernel_quantizer_internal, kernel_constraint, kernel_initializer
        )

        if use_bias:
            bias_constraint, bias_initializer = get_auto_range_constraint_initializer(
                self.bias_quantizer_internal, bias_constraint, bias_initializer
            )
        if activation is not None:
            activation = get_quantizer(activation)

        super().__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs,
        )

    def call(self, inputs):
        if self.kernel_quantizer:
            quantized_kernel = self.kernel_quantizer_internal(self.kernel)
        else:
            quantized_kernel = self.kernel
        output = ops.dot(inputs, quantized_kernel)
        if self.use_bias:
            if self.bias_quantizer:
                quantized_bias = self.bias_quantizer_internal(self.bias)
            else:
                quantized_bias = self.bias
            output = tf.nn.bias_add(output, quantized_bias, data_format='N...C')
        if self.activation is not None:
            output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) >= 2
        assert input_shape[-1]
        output_shape = list(input_shape)
        output_shape[-1] = self.units
        return tuple(output_shape)

    def get_config(self):
        config = {
            'units': self.units,
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'kernel_quantizer': constraints.serialize(self.kernel_quantizer_internal),
            'bias_quantizer': constraints.serialize(self.bias_quantizer_internal),
            'kernel_initializer': initializers.serialize(self.kernel_initializer),
            'bias_initializer': initializers.serialize(self.bias_initializer),
            'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer': regularizers.serialize(self.bias_regularizer),
            'activity_regularizer': regularizers.serialize(self.activity_regularizer),
            'kernel_constraint': constraints.serialize(self.kernel_constraint),
            'bias_constraint': constraints.serialize(self.bias_constraint),
            'kernel_range': self.kernel_range,
            'bias_range': self.bias_range,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def get_quantization_config(self):
        return {
            'kernel_quantizer': str(self.kernel_quantizer_internal),
            'bias_quantizer': str(self.bias_quantizer_internal),
            'activation': str(self.activation),
            'units': str(self.units),
        }

    def get_quantizers(self):
        return self.quantizers

    def get_prunable_weights(self):
        return [self.kernel]


class quantized_sigmoid(BaseQuantizer):  # pylint: disable=invalid-name
    """Computes a quantized sigmoid to a number of bits.

    Attributes:
      bits: number of bits to perform quantization.
      symmetric: if true, we will have the same number of values for positive
        and negative numbers.
      use_real_sigmoid: if true, will use the sigmoid from Keras backend
      use_stochastic_rounding: if true, we perform stochastic rounding.

    Returns:
      Function that performs sigmoid + quantization to bits in the range 0.0 to 1.0.
    """

    def __init__(self, bits=8, symmetric=False, use_real_sigmoid=False, use_stochastic_rounding=False):
        super().__init__()
        self.bits = bits
        self.symmetric = symmetric
        self.use_real_sigmoid = use_real_sigmoid
        self.use_stochastic_rounding = use_stochastic_rounding

    def __str__(self):
        flags = [str(self.bits)]
        if self.symmetric:
            flags.append(str(int(self.symmetric)))
        if self.use_real_sigmoid:
            flags.append(str(int(self.use_real_sigmoid)))
        if self.use_stochastic_rounding:
            flags.append(str(int(self.use_stochastic_rounding)))
        return 'quantized_sigmoid(' + ','.join(flags) + ')'

    def __call__(self, x):
        x = ops.cast(x, 'float32')
        m = ops.power(2.0, self.bits)

        p = ops.sigmoid(x) if self.use_real_sigmoid else _sigmoid(x)

        return ops.clip((_round_through(p * m, self.use_stochastic_rounding) / m), (1.0 * self.symmetric) / m, 1.0 - 1.0 / m)

    def max(self):
        """Get the maximum value that quantized_sigmoid can represent."""
        return 1.0 - 1.0 / ops.power(2, self.bits)

    def min(self):
        """Get the minimum value that quantized_sigmoid can represent."""
        return (1.0 * self.symmetric) / ops.power(2, self.bits)

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def get_config(self):
        config = {
            'bits': self.bits,
            'symmetric': self.symmetric,
            'use_real_sigmoid': self.use_real_sigmoid,
            'use_stochastic_rounding': self.use_stochastic_rounding,
        }
        return config



import random

import keras
import tensorflow as tf
from keras.api import Model
from keras.api.callbacks import Callback


class FreeBOPs(Callback):
    def on_epoch_end(self, epoch, logs=None):
        assert logs is not None
        assert isinstance(self.model, Model)
        bops = 0
        for layer in self.model.layers:
            if hasattr(layer, 'compute_bops'):
                bops += layer.compute_bops()
        logs['bops'] = bops


def round(x, round_type):
    if round_type == 0:  # standard round
        return tf.floor(x + 0.5)
    if round_type == 1:  # stochastic round
        _floor = tf.floor(x)
        noise = tf.random.uniform(tf.shape(x))
        return tf.where(noise < x - _floor, _floor + 1, _floor)


def quantize(x, bits, round_type):
    """Quantize a tensor to a given number of bits."""
    scale = tf.pow(2.0, bits) - 1
    quantized = round(x * scale, round_type) / scale
    return x + tf.stop_gradient(quantized - x)  # STE approximation


def quantized_sigmoid(x, bits, round_type):
    """
    Quantized version of the sigmoid activation function.

    Args:
        x (tf.Tensor): Input tensor.
        bits (int): Number of bits to quantize the output to.

    Returns:
        tf.Tensor: Quantized sigmoid output.
    """
    scale = tf.pow(2.0, bits) - 1  # Number of discrete levels
    sigmoid = tf.math.sigmoid(x)
    quantized = round(sigmoid * scale, round_type) / scale  # Quantize to discrete levels
    return sigmoid + tf.stop_gradient(quantized - sigmoid)  # STE approximation


def quantized_relu(x, bits, max_value, round_type):
    """
    Quantized version of the ReLU activation function.

    Args:
        x (tf.Tensor): Input tensor.
        bits (int): Number of bits to quantize the output to.
        max_value (float): Maximum value for clipping (e.g., 6.0 for ReLU6).

    Returns:
        tf.Tensor: Quantized ReLU output.
    """
    scale = tf.pow(2.0, bits) - 1  # Number of discrete levels
    relu = tf.clip_by_value(x, 0.0, max_value)  # Clip to [0, max_value]
    quantized = round(relu * scale / max_value, round_type) * max_value / scale  # Quantize to discrete levels
    return relu + tf.stop_gradient(quantized - relu)  # STE approximation


class TQActivation(keras.layers.Layer):
    def __init__(
        self, input_shape, bits, min_bits=0, max_bits=32, clip_min=-1.0, clip_max=1.0, round_type=1, alpha='random', **kwargs
    ):
        """
        Trainable Quantized Activation Layer.

        Args:
            min_bits (int): Minimum bit width.
            max_bits (int): Maximum bit width.
            clip_min (float): Minimum value for clipping activations.
            clip_max (float): Maximum value for clipping activations.
        """
        super().__init__(**kwargs)
        self.input_shape = input_shape
        self.bits = bits
        if type(bits) is str and bits == 'random':
            self.bits = random.randrange(16, 32)
        self.min_bits = min_bits
        self.max_bits = max_bits
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.round_type = round_type
        self.alpha = alpha
        if type(alpha) is str and alpha == 'random':
            self.alpha = random.uniform(0, 1)

    def build(self, input_shape):
        # Trainable bit width
        self.activation_bits = self.add_weight(
            name='bits',
            shape=(input_shape[-1],),
            initializer=keras.initializers.Constant(self.bits),
            trainable=True,
        )

    def call(self, inputs):
        # Quantize the activations using the current bit width
        inputs = tf.clip_by_value(inputs, self.clip_min, self.clip_max)
        quantized_output = quantize(inputs, self.activation_bits, self.round_type)

        self.add_loss(self.alpha * self.compute_bops())
        return quantized_output

    def compute_output_shape(self, input_shape):
        # The output shape is the same as the input shape
        return input_shape

    def compute_bops(self):
        """
        Compute the BOPs for this layer.
        Layer BOPs = (num_activations x activation_bits)

        Args:
            num_activations (int): Number of activations in the layer output.

        Returns:
            float: BOPs for this layer.
        """

        activation_bops = tf.reduce_sum(self.activation_bits)

        # Total BOPs for the layer
        total_bops = activation_bops
        return total_bops

    def compute_bops_std(self):
        activation_bops_std = tf.math.reduce_std(self.activation_bits)
        total_bops_std = activation_bops_std
        return total_bops_std

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'min_bits': self.min_bits,
                'max_bits': self.max_bits,
                'clip_min': self.clip_min,
                'clip_max': self.clip_max,
            }
        )
        return config


class TQDense(keras.layers.Layer):
    def __init__(
        self,
        units,
        activation='linear',
        init_bits=16,
        min_bits=0,
        max_bits=32,
        max_value_relu=32,
        alpha='random',
        round_type=1,
        **kwargs,
    ):
        """
        A quantized Dense layer with trainable bit widths for weights, activations, and biases.

        Args:
            units (int): Number of output units.
            min_bits (int): Minimum bit width allowed.
            max_bits (int): Maximum bit width allowed.
            kwargs: Additional arguments for the Layer superclass.
        """
        super().__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.init_bits_weight = init_bits
        self.init_bits_activation = init_bits
        self.init_bits_bias = init_bits
        if type(init_bits) is str and init_bits == 'random':
            # TODO: set range of bits
            self.init_bits_weight = random.randrange(8, 32)
            self.init_bits_activation = random.randrange(8, 32)
            self.init_bits_bias = random.randrange(8, 32)
        self.min_bits = min_bits
        self.max_bits = max_bits
        self.max_value_relu = max_value_relu
        self.alpha = alpha
        if type(alpha) is str and alpha == 'random':
            self.alpha = random.uniform(0, 1)
        self.round_type = round_type

    def build(self, input_shape):
        # Initialize weights, biases, and their bit widths
        self.kernel = self.add_weight(
            name='kernel',
            shape=(input_shape[-1], self.units),
            initializer='he_normal',
            trainable=True,
        )
        self.bias = self.add_weight(
            name='bias',
            shape=(self.units,),
            initializer='zeros',
            trainable=True,
        )

        # Trainable bit widths
        self.weight_bits = self.add_weight(
            name='weight_bits',
            shape=(input_shape[-1], self.units),
            initializer=keras.initializers.Constant(self.init_bits_weight),
            trainable=True,
            dtype=tf.float32,
        )
        self.activation_bits = self.add_weight(
            name='activation_bits',
            shape=(self.units,),
            initializer=keras.initializers.Constant(self.init_bits_activation),
            trainable=True,
            dtype=tf.float32,
        )
        self.bias_bits = self.add_weight(
            name='bias_bits',
            shape=(self.units,),
            initializer=keras.initializers.Constant(self.init_bits_bias),
            trainable=True,
            dtype=tf.float32,
        )

    def compute_bops(self):
        """
        Compute the BOPs for this layer.
        Layer BOPs = (num_weights x weight_bits) + (num_activations x activation_bits) + (num_biases x bias_bits)

        Args:
            num_activations (int): Number of activations in the layer output.

        Returns:
            float: BOPs for this layer.
        """

        # Compute individual weight BOPs
        weight_bops = tf.reduce_sum(self.weight_bits)  # Sum trainable bit-widths per weight
        bias_bops = tf.reduce_sum(self.bias_bits)
        activation_bops = tf.reduce_sum(self.activation_bits)

        # Total BOPs for the layer
        total_bops = weight_bops + bias_bops + activation_bops
        return total_bops

    def compute_bops_std(self):
        weight_bops_std = tf.math.reduce_std(self.weight_bits)
        bias_bops_std = tf.math.reduce_std(self.bias_bits)
        activation_bops_std = tf.math.reduce_std(self.activation_bits)
        total_bops_std = [weight_bops_std, bias_bops_std, activation_bops_std]
        return tf.reduce_mean(total_bops_std)

    def call(self, inputs):
        # Quantize kernel (weights)
        unsigned_weight_bits = tf.clip_by_value(self.weight_bits, self.min_bits, self.max_bits)
        quantized_kernel = quantize(self.kernel, unsigned_weight_bits, self.round_type)

        # Compute the output
        output = tf.matmul(inputs, quantized_kernel)

        # Quantize bias
        unsigned_bias_bits = tf.clip_by_value(self.bias_bits, self.min_bits, self.max_bits)
        quantized_bias = quantize(self.bias, unsigned_bias_bits, self.round_type)
        output = tf.nn.bias_add(output, quantized_bias)

        # Quantize activations
        unsigned_activation_bits = tf.clip_by_value(self.activation_bits, self.min_bits, self.max_bits)
        if self.activation == 'sigmoid':
            output = quantized_sigmoid(output, unsigned_activation_bits, self.round_type)

        if self.activation == 'relu':
            output = quantized_relu(output, unsigned_activation_bits, self.max_value_relu, self.round_type)

        if self.activation == 'linear':
            output = quantize(output, unsigned_activation_bits, self.round_type)

        self.add_loss(self.alpha * self.compute_bops())

        return tf.convert_to_tensor(output)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'units': self.units,
                'min_bits': self.min_bits,
                'max_bits': self.max_bits,
            }
        )
        return config



from tensorflow import keras
from tensorflow.keras.models import load_model

import numpy as np

import hls4ml
import hls4ml.utils
import hls4ml.converters

from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from sklearn.preprocessing import MinMaxScaler

# Register the optimization passes (if any)
backend = hls4ml.backends.get_backend('Vitis')

# load data
normal_data = np.load('./../data/normal_data.npy', allow_pickle=True)
abnormal_data = np.load('./../data/abnormal_data.npy', allow_pickle=True)
nPV_normal = normal_data[:, 0]
nPV_abnormal = abnormal_data[:, 0]
normal_data = normal_data[:, 1:]
abnormal_data = abnormal_data[:, 1:]

y = np.concatenate((np.ones(len(abnormal_data)), np.zeros(len(normal_data))))
X = np.concatenate((abnormal_data, normal_data))
S = np.concatenate((nPV_abnormal, nPV_normal))

X_train, X_test, y_train, y_test, S_train, S_test = train_test_split(X, y, S, random_state=42)
scaler = MinMaxScaler()  # add robast scaler
X_train = scaler.fit_transform(X_train)
X_test = scaler.transform(X_test)

# keras model
input_layer = keras.layers.Input((X_train.shape[1],))
x = keras.layers.Dense(64, activation="relu")(input_layer)
x = keras.layers.Dense(32, activation="relu")(x)
x = keras.layers.Dense(16, activation="relu")(x)
output_layer = keras.layers.Dense(1, activation="sigmoid")(x)

# Create the model
model = keras.Model(inputs=input_layer, outputs=output_layer)

# Compile the model
model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['accuracy'])
model.fit(X_train, y_train, batch_size=256, epochs=1)

# test the model
y_pred = model.predict(X_test)
auc_value = roc_auc_score(y_test, y_pred)
print("Kears model: ", auc_value)

# save the model
model.save('../trained_models/test_keras.h5')

# load the model
model = load_model("../trained_models/test_keras.h5")

# test the model
y_pred = model.predict(X_test)
auc_value = roc_auc_score(y_test, y_pred)
print("Kears model: ", auc_value)

hmodel = hls4ml.converters.convert_from_keras_model(
    model,
    backend='Vitis',
    output_dir='testkeras-HLS4ML-V3'
)

# test the HLS4ML model
hmodel.compile()
X_test = np.ascontiguousarray(X_test)
y_pred = hmodel.predict(X_test)
auc_value = roc_auc_score(y_test, y_pred)
print("HLS4ML model: ", auc_value)

# build the model
hmodel.build()

hls4ml.report.read_vivado_report('testkeras-HLS4ML-V3')



import hls4ml
import hls4ml.utils
import hls4ml.converters


# setup custom layer
rev_config_template = """struct config{index} : nnet::bernoulli_config {{
    static const unsigned n_in = {n_in};
    const ap_ufixed<1,0> thr = 0.5;
}};\n"""

rev_function_template = 'nnet::bernoulli<{input_t}, {config}>({input}, {output});'
rev_include_list = ['nnet_utils/bernoulli.h']

class HBernoulli(hls4ml.model.layers.Layer):
    '''hls4ml implementation of the bernoulli layer'''

    def initialize(self):
        inp = self.get_input_variable()
        shape = inp.shape
        dims = inp.dim_names
        self.add_output_variable(shape, dims)


class HBernoulliConfigTemplate(hls4ml.backends.template.LayerConfigTemplate):
    def __init__(self):
        super().__init__(HBernoulli)
        self.template = rev_config_template

    def format(self, node):
        params = self._default_config_params(node)
        return self.template.format(**params)


class HBernoulliFunctionTemplate(hls4ml.backends.template.FunctionCallTemplate):
    def __init__(self):
        super().__init__(HBernoulli, include_header=rev_include_list)
        self.template = rev_function_template

    def format(self, node):
        params = self._default_function_params(node)
        return self.template.format(**params)


# Parser for converter
def parse_bernoulli_layer(keras_layer, input_names, input_shapes, data_reader, thr=0.5):
    layer = {}
    layer['class_name'] = 'BernoulliSampling'
    layer['name'] = 'bernoulli'#keras_layer['config']['name']
    layer['n_in'] = input_shapes[0][1]
    layer['thr'] = thr

    if input_names is not None:
        layer['inputs'] = input_names

    return layer, [shape for shape in input_shapes[0]]



import keras
import tensorflow as tf


@tf.custom_gradient
def grad_reverse(x):
    y = tf.identity(x)

    def custom_grad(dy):
        return -dy

    return y, custom_grad


class GradReverse(keras.layers.Layer):
    def __init__(self):
        super().__init__()

    def call(self, x):
        return grad_reverse(x)
